# -*- encoding:utf-8 -*-
"""
    资金模块，不区分美元，人民币等类型，做美股交易默认当作美元，a股默认当作人民币
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import logging

import numpy as np
import pandas as pd

from ..UtilBu.ABuProgress import AbuProgress
from ..UtilBu import ABuDateUtil
from ..TradeBu.ABuOrder import AbuOrder
from ..TradeBu.ABuCommission import AbuCommission
from ..CoreBu.ABuBase import PickleStateMixin

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuCapital(PickleStateMixin):
    """资金类"""

    def __init__(self, init_cash, benchmark, user_commission_dict=None):
        """
        :param init_cash: 初始资金值，注意这里不区分美元，人民币等类型，做美股交易默认当作美元，a股默认当作人民币，int
        :param benchmark: 资金回测时间标尺，做为资金类表格时间范围确定使用，AbuBenchmark实例对象
        :param user_commission_dict: dict对象，可自定义交易手续费计算方法，详情查看AbuCommission
        """
        self.read_cash = init_cash

        kl_pd = benchmark.kl_pd
        if kl_pd is None:
            # 要求基准必须有数据
            raise ValueError('CapitalClass init klPd is None')

        # 根据基准时间序列，制作相同的时序资金对象capital_pd(pd.DcataFrame对象)
        self.capital_pd = pd.DataFrame(
            {
                'cash_blance': np.NAN * kl_pd.shape[0],
                'stocks_blance': np.zeros(kl_pd.shape[0]),
                'atr21': kl_pd['atr21'],
                'date': kl_pd['date']
            },
            index=kl_pd.index)

        self.capital_pd['date'] = self.capital_pd['date'].astype(int)
        # cash_blance除了第一个其它都是nan
        self.capital_pd.loc[
            self.capital_pd.index[0], 'cash_blance'] = self.read_cash
        # 构造交易手续费对象AbuCommission，如果user自定义手续费计算方法，通过user_commission_dict传入
        self.commission = AbuCommission(user_commission_dict)

    def __str__(self):
        """打印对象显示：capital_pd.info commission_df.info"""
        return 'capital_pd:\n{}\ncommission_pd:\n{}'.format(self.capital_pd.info(),
                                                            self.commission.commission_df.info())

    __repr__ = __str__

    def __len__(self):
        """对象长度：时序资金对象capital_pd的行数，即self.capital_pd.shape[0]"""
        return self.capital_pd.shape[0]

    def init_k_line(self, a_symbol):
        """
        每一个交易对象在时序资金对象capital_pd上都添加对应的call keep（买涨持仓量），call worth（买涨总价值），
        put keep（买跌持仓量），put worth（买跌总价值）
        :param a_symbol: symbol str对象
        """

        # 买涨持仓量
        call_keep = '_call_keep'
        if self.capital_pd.columns.tolist().count(a_symbol + call_keep) == 0:
            self.capital_pd[a_symbol + call_keep] = np.NAN * \
                                                    self.capital_pd.shape[0]
        # 买跌持仓量
        put_keep = '_put_keep'
        if self.capital_pd.columns.tolist().count(a_symbol + put_keep) == 0:
            self.capital_pd[a_symbol + put_keep] = np.NAN * \
                                                   self.capital_pd.shape[0]
        # 买涨总价值
        call_worth = '_call_worth'
        if self.capital_pd.columns.tolist().count(a_symbol + call_worth) == 0:
            self.capital_pd[a_symbol + call_worth] = np.NAN * \
                                                     self.capital_pd.shape[0]

        # 买跌总价值
        put_worth = '_put_worth'
        if self.capital_pd.columns.tolist().count(a_symbol + put_worth) == 0:
            self.capital_pd[a_symbol + put_worth] = np.NAN * \
                                                    self.capital_pd.shape[0]

    def apply_init_kl(self, action_pd, show_progress):
        """
        根据回测交易在时序资金对象capital_pd上新建对应的call，put列
        :param action_pd: 回测交易行为对象，pd.DataFrame对象
        :param show_progress: 外部设置是否需要显示进度条
        """
        # 使用set筛选唯一的symbol交易序列
        symbols = set(action_pd.symbol)
        # 单进程进度条
        with AbuProgress(len(symbols), 0, label='apply_init_kl...') as progress:
            for pos, symbol in enumerate(symbols):
                if show_progress:
                    progress.show(a_progress=pos + 1)
                # 迭代symbols，新建对应的call，put列
                self.init_k_line(symbol)

    def apply_k_line(self, a_k_day, kl_pd, buy_type_head):
        """
        在apply_kl中的do_apply_kl方法中时序资金对象capital进行apply的对应方法，
        即迭代金融时间序列的每一个交易日，根据持仓量计算每一个交易日的市场价值
        :param a_k_day: 每一个被迭代中的时间，即每一个交易日数据
        :param kl_pd: 正在被apply迭代的金融时间序列本体，pd.DataFrame对象
        :param buy_type_head: 代表交易类型，范围（_call，_put）
        :return:
        """
        if a_k_day[kl_pd.name + buy_type_head + '_keep'] > 0:
            kl = kl_pd[kl_pd['date'] == a_k_day['date']]
            if kl is None or kl.shape[0] == 0:
                # 前提是当前交易日有对应的持仓
                return

            # 今天的收盘价格
            td_close = kl['close'].values[0]
            if buy_type_head == '_put':
                # 针对buy put对价格进行转换，主要是针对put特殊处理
                today_key = kl.key.values[0]
                if today_key > 0:
                    yd_close = kl_pd.iloc[today_key - 1].close
                    # 如果是买跌，实时市场收益以昨天为基础进行计算，即＊－1进行方向计算，以收益来映射重新定义今天的收盘价格
                    td_close = (td_close - yd_close) * -1 + yd_close
            # 根据持仓量即处理后的今日收盘价格，进行今日价值计算
            self.capital_pd.loc[kl.index, [kl_pd.name + buy_type_head + '_worth']] \
                = np.round(td_close * a_k_day[kl_pd.name + buy_type_head + '_keep'], 3)

    def apply_kl(self, action_pd, kl_pd_manager, show_progress):
        """
        apply_action之后对实际成交的交易分别迭代更新时序资金对象capital_pd上每一个交易日的实时价值
        :param action_pd: 回测结果生成的交易行为构成的pd.DataFrame对象
        :param kl_pd_manager: 金融时间序列管理对象，AbuKLManager实例
        :param show_progress: 是否显示进度条
        """

        # 在apply_action之后形成deal列后，set出考虑资金下成交了的交易序列
        deal_symbols_set = set(action_pd[action_pd['deal'] == 1].symbol)

        def do_apply_kl(kl_pd, buy_type_head):
            """
            根据金融时间序列在时序资金对象capital_pd上进行call（买涨），put（买跌）的交易日实时价值更新
            :param kl_pd: 金融时间序列，pd.DataFrame对象
            :param buy_type_head: 代表交易类型，范围（_call，_put）
            """
            # cash_blance对na进行pad处理
            self.capital_pd['cash_blance'].fillna(method='pad', inplace=True)
            # symbol对应列持仓量对na进行处理
            self.capital_pd[kl_pd.name + buy_type_head + '_keep'].fillna(method='pad', inplace=True)
            self.capital_pd[kl_pd.name + buy_type_head + '_keep'].fillna(0, inplace=True)

            # 使用apply在axis＝1上，即每一个交易日上对持仓量及市场价值进行更新
            self.capital_pd.apply(self.apply_k_line, axis=1, args=(kl_pd, buy_type_head))

            # symbol对应列市场价值对na进行处理
            self.capital_pd[kl_pd.name + buy_type_head + '_worth'].fillna(method='pad', inplace=True)
            self.capital_pd[kl_pd.name + buy_type_head + '_worth'].fillna(0, inplace=True)

            # 纠错处理把keep=0但是worth被pad的进行二次修正
            fe_mask = (self.capital_pd[kl_pd.name + buy_type_head + '_keep'] == 0) & (
                self.capital_pd[kl_pd.name + buy_type_head + '_worth'] > 0)
            # 筛出需要纠错的index
            fe_index = self.capital_pd[fe_mask].index
            # 将需要纠错的对应index上市场价值进行归零处理
            cp_w = self.capital_pd[kl_pd.name + buy_type_head + '_worth']
            cp_w.loc[fe_index] = 0

        # 单进程进度条
        with AbuProgress(len(deal_symbols_set), 0, label='apply_kl...') as progress:
            for pos, deal_symbol in enumerate(deal_symbols_set):
                if show_progress:
                    progress.show(a_progress=pos + 1)
                # 从kl_pd_manager中获取对应的金融时间序列kl，每一个kl分别进行call（买涨），put（买跌）的交易日实时价值更新
                kl = kl_pd_manager.get_pick_time_kl_pd(deal_symbol)
                # 进行call（买涨）的交易日实时价值更新
                do_apply_kl(kl, '_call')
                # 进行put（买跌）的交易日实时价值更新
                do_apply_kl(kl, '_put')

    def apply_action(self, a_action, progress):
        """
        在回测结果生成的交易行为构成的pd.DataFrame对象上进行apply对应本方法，即
        将交易行为根据资金情况进行处理，处理手续费以及时序资金对象capital_pd上的
        数据更新
        :param a_action: 每一个被迭代中的action，即每一个交易行为
        :param progress: 进度条对象
        :return: 是否成交deal bool
        """
        # 区别买入行为和卖出行为
        is_buy = True if a_action['action'] == 'buy' else False
        # 从action数据构造AbuOrder对象
        order = AbuOrder()
        order.buy_symbol = a_action['symbol']
        order.buy_cnt = a_action['Cnt']
        if is_buy:
            # 如果是买单，sell_price = price2 ,详情阅读ABuTradeExecute中transform_action
            order.buy_price = a_action['Price']
            order.sell_price = a_action['Price2']
        else:
            # 如果是卖单，buy_price = price2 ,详情阅读ABuTradeExecute中transform_action
            order.sell_price = a_action['Price']
            order.buy_price = a_action['Price2']
        # 交易发生的时间
        order.buy_date = a_action['Date']
        order.sell_date = a_action['Date']
        # 交易的方向
        order.expect_direction = a_action['Direction']

        # 对买单和卖单分别进行处理，确定是否成交deal
        deal = self.buy_stock(order) if is_buy else self.sell_stock(order)

        if progress is not None:
            progress.show()

        return deal

    def buy_stock(self, a_order):
        """
        在apply_action中每笔交易进行处理，根据买单计算cost，在时序资金对象capital_pd上修改对应cash_blance，
        以及更新对应symbol上的持仓量
        :param a_order: 在apply_action中由action转换的AbuOrder对象
        :return: 是否成交deal bool
        """

        # 首先使用commission对象计算手续费
        with self.commission.buy_commission_func(a_order) as (buy_func, commission_list):
            commission = buy_func(a_order.buy_cnt, a_order.buy_price)
            # 将上下文管理器中返回的commission_list中添加计算结果commission，内部根据list长度决定写入手续费记录pd.DataFrame
            commission_list.append(commission)
        # cost = 买单数量 ＊ 单位价格 ＋ 手续费
        order_cost = a_order.buy_cnt * a_order.buy_price + commission
        # 买单时间转换成pd时间日期对象
        time_ind = pd.to_datetime(ABuDateUtil.fmt_date(a_order.buy_date))
        # pd时间日期对象置换出对应的index number
        num_index = self.capital_pd.index.tolist().index(time_ind)

        # cash_blance初始化init中除了第一个其它都是nan
        cash_blance = self.capital_pd['cash_blance'].dropna()

        # 先截取cash_blance中 < time_ind的时间序列
        bl_ss = cash_blance[cash_blance.index <= time_ind]
        if bl_ss is None or bl_ss.shape[0] == 0:
            logging.info('bl_ss.shape[0] == 0 ' + str(a_order.buy_date))
            return False
        # 截取的bl_ss中选中最后一个元素做为判定买入时刻的cash值
        cash = bl_ss.iloc[-1]
        # 判定买入时刻的cash值是否能够钱买入
        if cash >= order_cost and a_order.buy_cnt > 0:
            # 够的话，买入，先将cash － cost
            cash -= order_cost
            # 根据a_order.expect_direction确定是要更新call的持仓量还是put的持仓量
            buy_type_keep = '_call_keep' if a_order.expect_direction == 1.0 else '_put_keep'
            # 前提1: 资金时间序列中有这个a_order.buy_symbol + buy_type_keep列
            has_cond1 = self.capital_pd.columns.tolist().count(a_order.buy_symbol + buy_type_keep) > 0
            # 前提2: 对应这个列从单子买入日期index开始，在dropna后还有shape说明本就有持仓
            has_cond2 = self.capital_pd[a_order.buy_symbol +
                                        buy_type_keep].iloc[:num_index + 1].dropna().shape[0] > 0

            keep_cnt = 0
            if has_cond1 and has_cond2:
                # 前提1 + 前提2->本就有持仓, 拿到之前的持仓量
                keep_cnt = self.capital_pd[a_order.buy_symbol
                                           + buy_type_keep].iloc[:num_index + 1].dropna()[-1]

            keep_cnt += a_order.buy_cnt

            # TODO 这里迁移之前逻辑，删除了一些未迁移模块的逻辑，所以看起来多写了一个流程，需要重构逻辑
            # 将计算好的cash更新到资金时间序列中对应的位置
            self.capital_pd.loc[time_ind, ['cash_blance']] = np.round(cash, 3)
            # 资金时间序列中之后的cash_blance也对应开始减去order_cost
            self.capital_pd.loc[cash_blance[cash_blance.index > time_ind].index, ['cash_blance']] \
                -= order_cost
            # 在更新持仓量前，取出之前的数值
            org_cnt = self.capital_pd.loc[time_ind][a_order.buy_symbol + buy_type_keep]
            # 更新持仓量
            self.capital_pd.loc[time_ind, [a_order.buy_symbol + buy_type_keep]] = keep_cnt
            if not np.isnan(org_cnt):
                # 对多个因子作用在同一个symbol上，且重叠了持股时间提前更新之后的持仓量
                keep_pos = self.capital_pd[a_order.buy_symbol + buy_type_keep].dropna()
                self.capital_pd.loc[keep_pos[keep_pos.index > time_ind].index, [a_order.buy_symbol + buy_type_keep]] \
                    += a_order.buy_cnt
            return True
        else:
            return False

    def sell_stock(self, a_order):
        """
        在apply_action中每笔交易进行处理，根据卖单计算cost，在时序资金对象capital_pd上修改对应cash_blance，
        以及更新对应symbol上的持仓量
        :param a_order: 在apply_action中由action转换的AbuOrder对象
        :return: 是否成交deal bool
        """

        # 卖单时间转换成pd时间日期对象
        time_ind = pd.to_datetime(ABuDateUtil.fmt_date(a_order.sell_date))
        # # pd时间日期对象置换出对应的index number
        num_index = self.capital_pd.index.tolist().index(time_ind)
        # 根据a_order.expect_direction确定是要更新call的持仓量还是put的持仓量
        buy_type_keep = '_call_keep' if a_order.expect_direction == 1.0 else '_put_keep'
        # 前提1: 资金时间序列中有这个a_order.buy_symbol + buy_type_keep列
        has_cond1 = self.capital_pd.columns.tolist().count(a_order.buy_symbol + buy_type_keep) > 0
        # 前提2: 对应这个列从单子买入日期index开始，在dropna后还有shape说明本就有持仓
        has_cond2 = self.capital_pd[a_order.buy_symbol + buy_type_keep].iloc[:num_index + 1].dropna().shape[0] > 0

        if has_cond1 and has_cond2:
            # 有持仓, 拿到之前的持仓量
            keep_cnt = self.capital_pd[a_order.buy_symbol + buy_type_keep].iloc[:num_index + 1].dropna()[-1]
            sell_cnt = a_order.buy_cnt

            if keep_cnt < sell_cnt:
                # 忽略一个问题，就算是当时买时的这个单子没有成交，这里也试图卖出当时设想买入的股数
                sell_cnt = keep_cnt
            if sell_cnt == 0:
                # 有可能由于没买入的单子，造成没有成交
                return False
            # 更新对应持仓量
            keep_cnt -= sell_cnt
            # 将卖出价格转换成call，put都可计算收益的价格，不要进行计算公式合并，保留冗余，便于理解
            sell_earn_price = (a_order.sell_price - a_order.buy_price) * a_order.expect_direction + a_order.buy_price
            order_earn = sell_earn_price * sell_cnt

            # 使用commission对象计算手续费
            with self.commission.sell_commission_func(a_order) as (sell_func, commission_list):
                commission = sell_func(sell_cnt, a_order.sell_price)
                # 将上下文管理器中返回的commission_list中添加计算结果commission，内部根据list长度决定写入手续费记录pd.DataFrame
                commission_list.append(commission)

            # cash_blance初始化init中除了第一个其它都是nan
            cash_blance = self.capital_pd['cash_blance'].dropna()
            # 截取 < time_indd的cash_blance中最后一个元素做为cash值
            cash = cash_blance[cash_blance.index <= time_ind].iloc[-1]
            cash += (order_earn - commission)
            # 将计算好的cash更新到资金时间序列中对应的位置
            self.capital_pd.loc[time_ind, ['cash_blance']] = np.round(cash, 3)

            # TODO 这里迁移之前逻辑，删除了一些未迁移模块的逻辑，所以看起来多写了一个流程，需要重构逻辑
            self.capital_pd.loc[cash_blance[cash_blance.index > time_ind].index, ['cash_blance']] \
                += (order_earn - commission)

            org_cnt = self.capital_pd.loc[time_ind][a_order.buy_symbol + buy_type_keep]
            # 更新持仓量
            self.capital_pd.loc[time_ind, [a_order.buy_symbol + buy_type_keep]] = keep_cnt

            if not np.isnan(org_cnt):
                # 针对diff factor same stock的情况
                keep_pos = self.capital_pd[a_order.buy_symbol + buy_type_keep].dropna()
                self.capital_pd.loc[keep_pos[keep_pos.index > time_ind].index, [a_order.buy_symbol + buy_type_keep]] \
                    -= sell_cnt
            return True
        else:
            return False



# coding=utf-8
"""
    对数据采集进行存储，读取，以及数据更新merge策略等实现模块
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from fnmatch import fnmatch

import pandas as pd

from ..CoreBu.ABuEnv import EDataCacheType, EMarketTargetType, EMarketSubType
from ..CoreBu import ABuEnv
from ..UtilBu.ABuFileUtil import load_df_csv, load_hdf5, ensure_dir, file_exist, del_file, dump_df_csv, \
    dump_del_hdf5
# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import xrange, range, filter
from ..UtilBu.ABuProgress import AbuProgress

try:
    from tables import HDF5ExtError
except ImportError:
    class HDF5ExtError(RuntimeError):
        """如果没有HDF5环境只能使用其它存贮模式"""
        pass

# 模块加载时统一确保文件夹存在，不在函数内部ensure_dir
ensure_dir(ABuEnv.g_project_kl_df_data)


def _kl_unique_key(symbol, start, end):
    """
    通过symbol以及start, end信息生成数据存储唯一id
    :param symbol: Symbol对象
    :param start: str日期对象 eg 2015-02-14
    :param end: str日期对象 eg 2017-02-14
    :return: e.g : 'usTSLA_2015-02-14_2017-02-14'
    """
    return "{}_{}_{}".format(symbol.value, start, end)


def rm_data_from_symbol(symbol):
    """
    删除特定symbol对应的本地缓存数据
    :param symbol: Symbol对象
    :return:
    """

    # TODO 只实现了针对hdf5的数据删除，添加其它存储模式的数据删除
    target_hdf5 = ABuEnv.g_project_kl_df_data
    with pd.HDFStore(target_hdf5) as h5s:

        if symbol in h5s:
            ind_key = h5s[symbol].values[0]
            if ind_key in h5s:
                # 删除缓存实体数据
                del h5s[ind_key]
            # 删除缓存数据index
            del h5s[symbol]


def load_all_kline(want_df=True, market=None, all_market=False):
    """
    只针对hdf5模式下生效，根据参数want_df决定读取hdf5中所有的index symbol数据或者实体pd.DataFrame数据
    :param want_df: 是要实体pd.DataFrame数据还是索引symbol数据
    :param market: 默认None，如None则服从ABuEnv.g_market_target市场设置
    :param all_market: 默认False, 如果True则不过滤市场，即忽略market参数指定的市场
    :return:
    """
    if ABuEnv.g_data_cache_type != EDataCacheType.E_DATA_CACHE_HDF5:
        raise RuntimeError('only support hdf5 cache mode!')

    # noinspection PyProtectedMember
    target_hdf5 = ABuEnv.g_project_kl_df_data_example if ABuEnv._g_enable_example_env_ipython \
        else ABuEnv.g_project_kl_df_data

    # 如果是实体pd.DataFrame数据的key，key的长度最少要两个日期 8 ＊ 2 ＋ 市场前缀 ＋ 2  ＋ 最小symbol长 ＋ 2
    k_min_index_key_len = 20  # 8 * 2 + 2 + 2
    with pd.HDFStore(target_hdf5) as h5s:

        # 根据参数want_df使用k_min_index_key_len过滤出需要的key序列
        keys = list(filter(
            lambda p_key: len(p_key) >= k_min_index_key_len if want_df else len(p_key) < k_min_index_key_len,
            h5s.keys()))

        if not all_market:
            # 非所有市场，即需要根据market再次过滤
            if market is None:
                market = ABuEnv.g_market_target

            k_market_map = {EMarketTargetType.E_MARKET_TARGET_US: [EMarketTargetType.E_MARKET_TARGET_US.value],
                            EMarketTargetType.E_MARKET_TARGET_HK: [EMarketTargetType.E_MARKET_TARGET_HK.value],
                            EMarketTargetType.E_MARKET_TARGET_CN: [EMarketSubType.SZ.value, EMarketSubType.SH.value]}

            # 对应市场的head
            market_head_list = k_market_map[market]

            def filter_market_key(p_key):
                """检测p_key是否startswith对应市场market_head_list"""
                for mh in market_head_list:
                    # key[0] = '/'
                    if p_key[1:].startswith(mh):
                        return True
                return False
            # 筛选出指定市场的key
            keys = list(filter(lambda p_key: filter_market_key(p_key), keys))
        # 结果返回序列，序列元素由（(key, h5s[key]) 构成
        return [(key, h5s[key]) for key in keys]


def covert_hdf_to_csv():
    """转换hdf5下的所有cache缓存至csv文件存贮格式"""

    # 获取hdf5下所有数据
    dfs = load_all_kline(all_market=True)

    # 临时保存存贮模式
    tmp_cache = ABuEnv.g_data_cache_type
    ABuEnv.g_data_cache_type = EDataCacheType.E_DATA_CACHE_CSV
    with AbuProgress(len(dfs), 0, 'csv covert') as pg:
        for symbol, dump_df in dfs:
            pg.show()
            # eg: usTSLA
            symbol_key = symbol.split('_')[0][1:]
            # eg: usTSLA_20110808_20170808
            date_key = symbol[1:]
            # 将df转换为csv格式
            dump_kline_df(dump_df, symbol_key, date_key)
    # 还原之前的存贮模式
    ABuEnv.g_data_cache_type = tmp_cache


def load_kline_df(symbol_key):
    """
    封装不同存储模式，根据symbol_key读取对应的本地缓存金融时间序列对象数据
    :param symbol_key: str对象symbol
    :return: (金融时间序列pd.DataFrame对象，索引date_key中start请求日期int，索引date_key中end请求日期int)
    """

    # 初始化默认读取日k数据使用_load_kline_hdf5方法
    load_kline_func = _load_kline_hdf5
    # 初始化默认读取日k数据key使用_load_hdf5_key方法
    load_kline_key = _load_hdf5_key
    # noinspection PyProtectedMember
    if ABuEnv.g_data_cache_type == EDataCacheType.E_DATA_CACHE_CSV and \
            not ABuEnv._g_enable_example_env_ipython:
        # 读取方式是csv，并且不是沙盒数据模式，切换load_kline_func，load_kline_key为csv读取函数
        load_kline_func = _load_kline_csv
        load_kline_key = _load_csv_key

    # noinspection PyUnusedLocal
    date_key = None
    try:
        # 首先通过symbol_key查询对应的金融时间序列是否存在索引date_key
        date_key = load_kline_key(symbol_key)
    except HDF5ExtError as e:
        # r_s = False的话，hdf5物理性错误就删除了，重来，所以重要的hdf5需要手动备份.
        r_s = True
        raise RuntimeError('hdf5 load error!! err={} '.format(e)) if r_s else os.remove(ABuEnv.g_project_kl_df_data)

    if date_key is not None:
        # 在索引date_key存在的情况下，继续查询实体金融时间序列对象
        df = load_kline_func(date_key[0])
        if df is not None:
            df['key'] = list(range(0, len(df)))
            # 索引date_key中转换df_req_start
            df_req_start = int(date_key[0][-17: -9])
            # 索引date_key中转换df_req_end
            df_req_end = int(date_key[0][-8:])
            return df, df_req_start, df_req_end
    return None, 0, 0


def _load_kline_csv(date_key):
    """
    针对csv存储模式，读取本地cache金融时间序列
    :param date_key: 金融时间序列索引key，针对对csv存储模式为目标csv的具体文件名
    """

    # 通过连接date_key和csv存储根目录，得到目标csv文件路径
    csv_fn = os.path.join(ABuEnv.g_project_kl_df_data_csv, date_key)
    df = load_df_csv(csv_fn)
    # 这里要把类型转换为time
    df.index = pd.to_datetime(df.index)
    return df


def _load_kline_hdf5(date_key):
    """
    针对hdf5存储模式，读取本地cache金融时间序列
    :param date_key: 金融时间序列索引key，针对对hdf5存储模式为目标金融时间序列查询key
    """
    # noinspection PyProtectedMember
    target_hdf5 = ABuEnv.g_project_kl_df_data_example if ABuEnv._g_enable_example_env_ipython \
        else ABuEnv.g_project_kl_df_data
    return load_hdf5(target_hdf5, date_key)


def _load_csv_key(symbol_key):
    """
    针对csv存储模式，通过symbol_key字符串找到对应的csv具体文件名称，
    如从usTSLA->找到usTSLA_2014-7-26_2016_7_26这个具体csv文件路径
    :param symbol_key: str对象，eg. usTSLA
    """
    if file_exist(ABuEnv.g_project_kl_df_data_csv):
        for name in os.listdir(ABuEnv.g_project_kl_df_data_csv):
            # 从csv缓存文件夹下进行模糊查询通过fnmatch匹配具体csv文件路径，eg. usTSLA->usTSLA_2014-7-26_2016_7_26
            # if fnmatch(name, '{}*'.format(symbol_key)):
            """
                这里不能模糊匹配，否则会因为TSL匹配上TSLA导致删除原有的symbol
                而且必须要加'_'做为symbol结束匹配标记
            """
            if name.startswith(symbol_key + '_'):
                # []只是为了配合外面针对不同store统一使用key[0]
                return [name]
    return None


def _load_hdf5_key(symbol_key):
    """
    针对hdf5存储模式，通过symbol_key字符串找到对应的在hdf5中的实体金融时间
    序列实体的索引序列
    :param symbol_key: 金融时间序列索引key，针对对hdf5存储模式为目标金融时间序列查询key
    """
    # noinspection PyProtectedMember
    target_hdf5 = ABuEnv.g_project_kl_df_data_example if ABuEnv._g_enable_example_env_ipython \
        else ABuEnv.g_project_kl_df_data
    return load_hdf5(target_hdf5, symbol_key)


def dump_kline_df(dump_df, symbol_key, date_key):
    """
    封装不同存储模式，根据symbol_key，date_key存储dump_df金融时间序列
    储存方法 symbol_key->date_key->dump_df

    eg  : usTSLA->usTSLA_20100214_20170214->tsla_df

    :param dump_df: 需要存储的金融时间序列实体pd.DataFrame对象
    :param symbol_key: str对象，eg. usTSLA
    :param date_key: str对象，eg. usTSLA_20100214_20170214 包含了df的时间开始时间与结束时间，便于计算需要的数据段是否在此之间

    """
    # 默认hdf5模式分配工作函数
    load_kline_key = _load_hdf5_key
    dump_kline_func = _dump_kline_hdf5
    load_kline_func = _load_kline_hdf5
    # csv模式分配工作函数
    if ABuEnv.g_data_cache_type == EDataCacheType.E_DATA_CACHE_CSV:
        dump_kline_func = _dump_kline_csv
        load_kline_key = _load_csv_key
        load_kline_func = _load_kline_csv

    _start = int(date_key[-17: -9])
    _end = int(date_key[-8:])

    # 读取本地缓存symbol_key对应的索引对象df_date_key
    df_date_key = load_kline_key(symbol_key)
    if df_date_key is not None:
        # 即之前存在本地缓存，需要merge金融时间序列

        # 之前本地缓存的时间序列结束日期
        df_end = int(df_date_key[0][-8:])
        # 之前本地缓存的时间序列开始日期
        df_start = int(df_date_key[0][-17: -9])
        if _start <= df_start and df_end <= _end:
            """
                新请求回来的数据完全包裹了原来的数据，直接存储即可
                       _start o------------------o _end
                       df_start   o--------o df_end

            result:    _start o------------------o _end
            """
            dump_kline_func(symbol_key, date_key, dump_df, delete_key=df_date_key)
        elif _start < df_start and df_end >= _end:
            """
                新请求回来的数据开始时间比原来的要考前，结束时间没有原来的远
                       _start o------------------o _end
                       df_start   o-----------------o df_end

            result:    _start o---------------------o df_end
            """
            date_key = '{}_{}_{}'.format(symbol_key, _start, df_end)
            # 首先读取原来的金融时间序列
            h5_df = load_kline_func(df_date_key[0])
            # [_start, df_start)
            # 选取dump_df.date < df_start部分
            new_df = dump_df[dump_df.date < df_start]
            # concat连起来两部分
            new_df = pd.concat([new_df, h5_df])
            # 最终保存的为new_df
            dump_kline_func(symbol_key, date_key, new_df, delete_key=df_date_key)
        elif _start >= df_start and df_end < _end:
            """
                新请求回来的数据开始时间比原来的晚，但结束时间也比原来的晚
                            _start o------------------o _end
                       df_start o-------------o df_end

            result:    df_start o---------------------o _end
            """
            date_key = '{}_{}_{}'.format(symbol_key, df_start, _end)
            # 首先读取原来的金融时间序列
            h5_df = load_kline_func(df_date_key[0])
            # 选取dump_df.date > df_end部分
            new_df = dump_df[dump_df.date > df_end]
            # concat连起来两部分
            new_df = pd.concat([h5_df, new_df])
            # 最终保存的为new_df
            dump_kline_func(symbol_key, date_key, new_df, delete_key=df_date_key)
        else:
            # 完全一样，是不是应该直接保存了呢，类似更新操作，暂时什么也不做
            pass
    else:
        # 即之前不存在本地缓存，直接存储在本地即可
        dump_kline_func(symbol_key, date_key, dump_df)


# noinspection PyUnusedLocal
def _dump_kline_csv(symbol_key, date_key, dump_df, delete_key=None):
    """
    针对csv存储模式，根据symbol_key，date_key存储dump_df金融时间序列
    :param symbol_key: str对象，eg. usTSLA，对于csv模式不需要，只为保持接口统一
    :param date_key: str对象，eg. usTSLA_20100214_20170214，csv模式下为对应的文件名
    :param dump_df: 需要存储的金融时间序列实体pd.DataFrame对象
    :param delete_key: 是否有需要删除的csv文件
    :return:
    """
    # 先删后后写入
    if delete_key is not None:
        delete_key = delete_key[0]
        del_fn = os.path.join(ABuEnv.g_project_kl_df_data_csv, delete_key)
        if file_exist(del_fn):
            del_file(del_fn)

    csv_fn = os.path.join(ABuEnv.g_project_kl_df_data_csv, date_key)
    dump_df_csv(csv_fn, dump_df)


def _dump_kline_hdf5(symbol_key, date_key, dump_df, delete_key=None):
    """
    针对hdf5存储模式，根据symbol_key，date_key存储dump_df金融时间序列
    :param symbol_key: str对象，eg. usTSLA，hdf5模式下为数据的索引key
    :param date_key: str对象，eg. usTSLA_20100214_20170214，hdf5模式下为金融时间序列实体key
    :param dump_df: 需要存储的金融时间序列实体pd.DataFrame对象
    :param delete_key: 是否有需要删除的csv文件
    :return:
    """

    # 需要dump的dict
    dump_dict = {symbol_key: pd.Series(date_key), date_key: dump_df}
    # 需要delete的dict
    del_array = [symbol_key]
    if delete_key is not None:
        del_array.append(delete_key)
    """
        dump中的target_hdf5只能是ABuEnv.g_project_kl_df_data，即使在沙盒测试数据下，
        在dump_del_hdf5中分别迭代dump_dict，del_array进行操作，详阅dump_del_hdf5
    """
    dump_del_hdf5(ABuEnv.g_project_kl_df_data, dump_dict, del_array)


def save_kline_df(df, temp_symbol, start_int, end_int):
    """
    独立对外的保存kl数据接口
    :param df: 需要存储的金融时间序列实体pd.DataFrame对象
    :param temp_symbol: Symbbol对象
    :param start_int: 请求的开始日期int
    :param end_int: 请求的结束日期int
    :return:
    """
    if df is not None:
        # 通过emp_symbol, start_int, end_int拼接唯一保存df_key
        df_key = _kl_unique_key(temp_symbol, start_int, end_int)
        dump_kline_df(df, temp_symbol.value, df_key)


def load_kline_df_net(source, temp_symbol, n_folds, start, end, start_int, end_int, save):
    """
    通过网络请求数据源，获取temp_symbol以及参数时间日期对应的金融时间序列pd.DataFrame对象
    :param source: 数据源BaseMarket的子类，非实例化对象
    :param temp_symbol: Symbol类对象
    :param n_folds: 需要获取几年的回测数据，int
    :param start: 开始回测日期，str对象
    :param end: 结束回测日期，str对象
    :param start_int: 开始回测日期，int
    :param end_int: 结束回测日期，int
    :param save: 是否从网络成功获取数据后进行数据的保存
    """
    df = None
    # 实例化数据源对象
    data_source = source(temp_symbol)

    if data_source.check_support():
        # 通过数据源混入的SupportMixin类检测数据源是否支持temp_symbol对应的市场数据
        df = data_source.kline(n_folds=n_folds, start=start, end=end)

    if df is not None and save:
        """
            这里的start_int， end_int会记作下次读取的df_req_start, df_req_end，即就是没有完整的数据返回，也可通过索引匹配上，
            即如果今天刚刚请求了直到今天为止的数据，但是数据源没有返回到今天的数据，今天的还没有，但是由于记录了end_int为今天，所以
            再次发起请求时不会走网络，会从本地获取数据
        """
        df_key = _kl_unique_key(temp_symbol, start_int, end_int)
        dump_kline_df(df, temp_symbol.value, df_key)
    return df



# -*- encoding:utf-8 -*-
"""
    选股示例因子：价格选股因子
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from .ABuPickStockBase import AbuPickStockBase, reversed_result
from ..TLineBu.ABuTL import AbuTLine
from ..CoreBu.ABuEnv import EMarketDataSplitMode
from ..MarketBu import ABuSymbolPd
from ..TradeBu import AbuBenchmark

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuPickStockShiftDistance(AbuPickStockBase):
    """位移路程比选股因子示例类"""

    def _init_self(self, **kwargs):
        """通过kwargs设置位移路程比选股条件，配置因子参数"""
        self.threshold_sd = kwargs.pop('threshold_sd', 2.0)
        self.threshold_max_cnt = kwargs.pop('threshold_max_cnt', 4)
        self.threshold_min_cnt = kwargs.pop('threshold_min_cnt', 1)

    @reversed_result
    def fit_pick(self, kl_pd, target_symbol):
        """开始根据位移路程比边际参数进行选股"""

        pick_line = AbuTLine(kl_pd.close, 'shift distance')
        shift_distance = pick_line.show_shift_distance(step_x=1.2, show_log=False, show=False)
        shift_distance = np.array(shift_distance)
        # show_shift_distance返回的参数为四组数据，最后一组是每个时间段的位移路程比值
        sd_arr = shift_distance[:, -1]
        # 大于阀值的进行累加和计算
        # noinspection PyUnresolvedReferences
        threshold_cnt = (sd_arr >= self.threshold_sd).sum()
        # 边际条件参数开始生效
        if self.threshold_max_cnt > threshold_cnt >= self.threshold_min_cnt:
            return True
        return False

    def fit_first_choice(self, pick_worker, choice_symbols, *args, **kwargs):
        raise NotImplementedError('AbuPickStockShiftDistance fit_first_choice unsupported now!')


class AbuPickStockNTop(AbuPickStockBase):
    """根据一段时间内的涨幅选取top N个"""

    def _init_self(self, **kwargs):
        """通过kwargs设置选股条件，配置因子参数"""
        # 选股参数symbol_pool：进行涨幅比较的top n个symbol
        self.symbol_pool = kwargs.pop('symbol_pool', [])
        # 选股参数n_top：选取前n_top个symbol, 默认3
        self.n_top = kwargs.pop('n_top', 3)
        # 选股参数direction_top：选取前n_top个的方向，即选择涨的多的，还是选择跌的多的
        self.direction_top = kwargs.pop('direction_top', 1)

    @reversed_result
    def fit_pick(self, kl_pd, target_symbol):
        """开始根据参数进行选股"""
        if len(self.symbol_pool) == 0:
            # 如果没有传递任何参照序列symbol，择默认为选中
            return True
        # 定义lambda函数计算周期内change
        kl_change = lambda p_kl: \
            p_kl.iloc[-1].close / p_kl.iloc[0].close if p_kl.iloc[0].close != 0 else 0

        cmp_top_array = []
        kl_pd.name = target_symbol
        # AbuBenchmark直接传递一个kl
        benchmark = AbuBenchmark(benchmark_kl_pd=kl_pd)
        for symbol in self.symbol_pool:
            if symbol != target_symbol:
                # 使用benchmark模式进行获取
                kl = ABuSymbolPd.make_kl_df(symbol, data_mode=EMarketDataSplitMode.E_DATA_SPLIT_UNDO,
                                            benchmark=benchmark)
                # kl = ABuSymbolPd.make_kl_df(symbol, start=start, end=end)
                if kl is not None and kl.shape[0] > kl_pd.shape[0] * 0.75:
                    # 需要获取实际交易日数量，避免停盘等错误信号
                    cmp_top_array.append(kl_change(kl))

        if self.n_top > len(cmp_top_array):
            # 如果结果序列不足n_top个，直接认为选中
            return True

        # 与选股方向相乘，即结果只去top
        cmp_top_array = np.array(cmp_top_array) * self.direction_top
        # 计算本源的周期内涨跌幅度
        target_change = kl_change(kl_pd) * self.direction_top
        # sort排序小－》大, 非inplace
        cmp_top_array.sort()
        # [::-1]大－》小
        # noinspection PyTypeChecker
        if target_change > cmp_top_array[::-1][self.n_top - 1]:
            # 如果比排序后的第self.n_top位置上的大就认为选中
            return True
        return False

    def fit_first_choice(self, pick_worker, choice_symbols, *args, **kwargs):
        raise NotImplementedError('AbuPickStockNTop fit_first_choice unsupported now!')



# -*- encoding:utf-8 -*-
"""
    手续费模块
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import logging
from contextlib import contextmanager

import numpy as np
import pandas as pd

from ..MarketBu.ABuSymbolFutures import AbuFuturesCn
from ..CoreBu.ABuFixes import partial
from ..CoreBu import ABuEnv
from ..CoreBu.ABuEnv import EMarketTargetType
from .ABuOrder import OrderMarket
from ..MarketBu import ABuMarket

__author__ = '阿布'
__weixin__ = 'abu_quant'


# noinspection PyUnusedLocal
def calc_commission_us(trade_cnt, price):
    """
    美股计算交易费用：每股0.01，最低消费2.99
    :param trade_cnt: 交易的股数（int）
    :param price: 每股的价格（美元）（暂不使用，只是保持接口统一）
    :return: 计算结果手续费
    """
    # 每股手续费0.01
    commission = trade_cnt * 0.01
    if commission < 2.99:
        # 最低消费2.99
        commission = 2.99
    return commission


def calc_commission_cn(trade_cnt, price):
    """
    a股计算交易费用：印花税＋佣金： 印花税万3，佣金万2.5
    :param trade_cnt: 交易的股数（int）
    :param price: 每股的价格（人民币）
    :return: 计算结果手续费
    """
    cost = trade_cnt * price
    # 印花税万3，
    tax = cost * 0.0003
    # 佣金万2.5
    commission = cost * 0.00025
    # 佣金最低5
    commission = commission if commission > 5 else 5
    commission += tax
    return commission


def calc_commission_hk(trade_cnt, price):
    """
    h股计算交易费用：印花税＋佣金： 佣金千分之二，印花税千分之一
    :param trade_cnt: 交易的股数（int）
    :param price: 每股的价格（人民币）
    :return: 计算结果手续费
    """
    cost = trade_cnt * price
    # 印花税千分之一
    tax = cost * 0.001
    # 佣金千分之二，
    commission = cost * 0.002
    commission += tax
    return commission


def calc_commission_tc(trade_cnt, price):
    """
    币类计算交易费用：只简单计算手续费，双向都使用流通币计算手续费，不涉及卖出使用币类的手续计算，
    如需要更精确计算，请使用自定义计算费率，即在AbuCommission初始化中自定义计算手续费的方法
    :param trade_cnt: 交易的币个数（int）
    :param price: 每币的价格（人民币）
    :return: 计算结果手续费
    """
    cost = trade_cnt * price
    # 双向都使用流通币计算手续费，千分之2
    commission = cost * 0.002
    return commission


# noinspection PyUnusedLocal
def calc_commission_futures_cn(trade_cnt, price, symbol_name):
    """
    期货计算交易费用：首先查询对应商品单位交易量（每手单位数量），以及每手手续费，再计算对应手续费
    :param trade_cnt: 交易的单位数量（int）
    :param price: 买入的价格（暂不使用，只是保持接口统一）
    :param symbol_name: 商品查询symbol
    :return: 计算结果手续费
    """
    min_unit = 10
    commission_unit = 10
    # 查询商品期货的对应df
    q_df = AbuFuturesCn().query_symbol(symbol_name)
    if q_df is not None:
        # 每手单位数量
        min_unit = q_df.min_unit.values[0]
        # 每手交易手续费
        commission_unit = q_df.commission.values[0]
    commission = trade_cnt / min_unit * commission_unit
    return commission


def calc_options_us(trade_cnt, price):
    """
    美股期权：差别很大，最好外部自定义自己的计算方法，这里只简单按照0.0035计算
    :param trade_cnt: 交易的股数（int）
    :param price: 每股的价格（美元
    :return: 计算结果手续费
    """
    cost = trade_cnt * price
    # 美股期权各个券商以及个人方式差别很大，最好外部自定义计算方法，这里只简单按照0.0035计算
    commission = cost * 0.0035
    return commission


def calc_commission_futures_global(trade_cnt, price):
    """
    国际期货：差别很大，最好外部自定义自己的计算方法，这里只简单按照0.002计算
    :param trade_cnt: 交易的股数（int）
    :param price: 每股的价格（美元）
    :return: 计算结果手续费
    """
    cost = trade_cnt * price
    # 国际期货各个券商以及代理方式差别很大，最好外部自定义计算方法，这里只简单按照0.002计算
    commission = cost * 0.002
    return commission


class AbuCommission(object):
    """交易手续费计算，记录，分析类，在AbuCapital中实例化"""

    def __init__(self, commission_dict):
        """
        :param commission_dict: 代表用户自定义手续费计算dict对象，
                                key：buy_commission_func， 代表用户自定义买入计算费用方法
                                key：sell_commission_func，代表用户自定义卖出计算费用方法
        """
        self.commission_dict = commission_dict
        # 对象内部记录交易的pd.DataFrame对象，列设定
        self.df_columns = ['type', 'date', 'symbol', 'commission']
        # 构建手续费记录pd.DataFrame对象commission_df
        self.commission_df = pd.DataFrame(columns=self.df_columns)

    def __str__(self):
        """打印对象显示：如果有手续费记录，打印记录df，否则打印commission_df.info"""
        if self.commission_df.shape[0] == 0:
            return str(self.commission_df.info())
        return str(self.commission_df)

    __repr__ = __str__

    # noinspection PyMethodMayBeStatic
    def _commission_enter(self, a_order):
        """
        通过a_order对象进行交易对象市场类型转换，分配对应手续费计算方法
        :param a_order: 交易单对象AbuOrder实例
        :return:
        """

        # 如果使用env中统一设置，即不需要通过OrderMarket对单子查询市场，提高运行效率，详ABuMarket
        market = ABuEnv.g_market_target if ABuMarket.g_use_env_market_set \
            else OrderMarket(a_order).symbol_market
        # 不同的市场不同的计算手续费方法
        if market == EMarketTargetType.E_MARKET_TARGET_US:
            # 美股
            calc_commission_func = calc_commission_us
        elif market == EMarketTargetType.E_MARKET_TARGET_CN:
            # a股
            calc_commission_func = calc_commission_cn
        elif market == EMarketTargetType.E_MARKET_TARGET_HK:
            # h股
            calc_commission_func = calc_commission_hk
        elif market == EMarketTargetType.E_MARKET_TARGET_TC:
            # 币类
            calc_commission_func = calc_commission_tc
        elif market == EMarketTargetType.E_MARKET_TARGET_FUTURES_CN:
            # 期货
            calc_commission_func = partial(calc_commission_futures_cn, symbol_name=a_order.buy_symbol)
        elif market == EMarketTargetType.E_MARKET_TARGET_OPTIONS_US:
            # 美股期权
            calc_commission_func = calc_options_us
        elif market == EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL:
            # 国际期货
            calc_commission_func = calc_commission_futures_global
        else:
            raise TypeError('buy_stock market error!!!')
        return calc_commission_func

    @contextmanager
    def buy_commission_func(self, a_order):
        """
        外部用with as 返回的list中需要加入计算的最终结果，否则不进行内部交易费用记录
        :param a_order: 买单对象AbuOrder实例
        """
        if self.commission_dict is not None and 'buy_commission_func' in self.commission_dict:
            # 如果有自定义计算交易费的方法使用自定义的
            buy_func = self.commission_dict['buy_commission_func']
        else:
            buy_func = self._commission_enter(a_order)

        # 使用list因为是可变类型，需要将外面的结果带回来
        commission_list = list()
        yield buy_func, commission_list

        # 如果有外部有append，说明需要记录手续费，且执行计算成功
        if len(commission_list) == 1:
            commission = commission_list[0]
            # 将买单对象AbuOrder实例中的数据转换成交易记录需要的np.array对象
            record = np.array(['buy', a_order.buy_date, a_order.buy_symbol, commission]).reshape(1, 4)
            record_df = pd.DataFrame(record, columns=self.df_columns)
            self.commission_df = self.commission_df.append(record_df)
        else:
            logging.info('buy_commission_func calc error')

    @contextmanager
    def sell_commission_func(self, a_order):
        """
        外部用with as 返回的list中需要加入计算的最终结果，否则不进行内部交易费用记录
        :param a_order: 卖单对象AbuOrder实例
        """
        if self.commission_dict is not None and 'sell_commission_func' in self.commission_dict:
            # 如果有自定义计算交易费的方法使用自定义的
            sell_func = self.commission_dict['sell_commission_func']
        else:
            sell_func = self._commission_enter(a_order)
        # 使用list因为是可变类型，需要将外面的结果带回来
        commission_list = list()

        yield sell_func, commission_list

        if len(commission_list) == 1:
            commission = commission_list[0]
            # 将卖单对象AbuOrder实例中的数据转换成交易记录需要的np.array对象
            record = np.array(['sell', a_order.sell_date, a_order.buy_symbol, commission]).reshape(1, 4)
            record_df = pd.DataFrame(record, columns=self.df_columns)
            self.commission_df = self.commission_df.append(record_df)
        else:
            logging.info('sell_commission_func calc error!!!')



# -*- encoding:utf-8 -*-
"""
    委托工具模块
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
from functools import update_wrapper
from operator import attrgetter

from ..CoreBu.ABuFixes import signature

__author__ = '阿布'
__weixin__ = 'abu_quant'


def first_delegate_has_method(delegate, check_params=True):
    """
    装饰在类函数上，如果delegate有定义对应名称方法，优先使用delegate中的方法，否则使用被装饰的方法

        eg:

        class A:
        def a_func(self):
            print('a.a_func')

        class B:
            def __init__(self):
                self.a = A()

            @ABuDelegateUtil.first_delegate_has_method('a')
            def a_func(self):
                print('b.a_func')

        in: B().a_func()
        out: a.a_func

    :param delegate: str对象，被委托的类属性对象名称，从被装饰的方法的类成员变量中寻找对应名字的对象
    :param check_params: 是否检测方法签名是否相同，默认检测
    """

    def decorate(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if len(args) > 0:
                wrap_self = args[0]
                # 首先从被装饰的类对象实例获取delegate
                delegate_obj = getattr(wrap_self, delegate, None)
                # 被装饰的类对象实例存在且存在和func.__name__一样的方法
                if delegate_obj is not None and hasattr(delegate_obj, func.__name__):
                    # 被委托的对象的函数的方法签名
                    delegate_params = list(signature(getattr(delegate_obj, func.__name__)).parameters.keys())
                    # 被装饰的函数的方法签名
                    func_params = list(signature(func).parameters.keys())[1:]
                    # print(func_params)
                    # print(delegate_params)
                    # TODO 增加检测规范，如只检测参数order及类型，不全匹配名字
                    if not check_params or delegate_params == func_params:
                        # 一致就优先使用被委托的对象的同名函数
                        return getattr(delegate_obj, func.__name__)(*args[1:], **kwargs)
            return func(*args, **kwargs)

        return wrapper

    return decorate


def replace_word_delegate_has_method(delegate, key_word, replace_word, check_params=True):
    """
    不在delegate中寻找完全一样的方法名字，在被装饰的方法名字中的key_word替换为replace_word后再在delegate中寻找，找到优先使用
    否则继续使用被装饰的方法
        eg:
            class A:
                def a_func(self):
                    print('a.a_func')

            class B:
                def __init__(self):
                    self.a = A()

                @ABuDelegateUtil.replace_word_delegate_has_method('a', key_word='b', replace_word='a')
                def b_func(self):
                    print('b.b_func')

            in: B().b_func()
            out: a.a_func

    :param delegate: str对象，被委托的类属性对象名称，从被装饰的方法的类成员变量中寻找对应名字的对象
    :param key_word: 被装饰的函数名称中将被replace_word替换的key_word，str对象
    :param replace_word: 替换key_word形成要寻找的在被委托函数中的名字，str对象
    :param check_params: 是否检测方法签名是否相同，默认检测
    """

    def decorate(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if len(args) > 0:
                wrap_self = args[0]
                # 首先从被装饰的类对象实例获取delegate
                delegate_obj = getattr(wrap_self, delegate, None)
                if delegate_obj is not None:
                    #  被装饰的类对象实例存在
                    org_func_name = func.__name__
                    if len(replace_word) > 0 and key_word in org_func_name:
                        # 使用replace_word替换原始函数名称org_func_name中的key_word
                        delegate_func_name = org_func_name.replace(key_word, replace_word)
                    else:
                        delegate_func_name = org_func_name

                    if hasattr(delegate_obj, delegate_func_name):
                        # 被装饰的类对象中确实存在delegate_func_name
                        delegate_params = list(signature(getattr(delegate_obj, delegate_func_name)).parameters.keys())
                        func_params = list(signature(func).parameters.keys())[1:]
                        # TODO 增加检测规范，如只检测参数order及类型，不全匹配名字
                        if not check_params or delegate_params == func_params:
                            # 参数命名一致就优先使用被委托的对象的函数
                            return getattr(delegate_obj, delegate_func_name)(*args[1:], **kwargs)
            return func(*args, **kwargs)

        return wrapper

    return decorate


class _IffHasAttrDescriptor(object):
    """
        摘自sklearn中metaestimators.py _IffHasAttrDescriptor
    """

    def __init__(self, fn, delegate_names, attribute_name):
        self.fn = fn
        self.delegate_names = delegate_names
        self.attribute_name = attribute_name

        # update the docstring of the descriptor
        update_wrapper(self, fn)

    def __get__(self, obj, p_type=None):
        # raise an AttributeError if the attribute is not present on the object
        if obj is not None:
            # delegate only on instances, not the classes.
            # this is to allow access to the docstrings.
            for delegate_name in self.delegate_names:
                try:
                    delegate = attrgetter(delegate_name)(obj)
                except AttributeError:
                    continue
                else:
                    getattr(delegate, self.attribute_name)
                    break
            else:
                attrgetter(self.delegate_names[-1])(obj)

        # lambda, but not partial, allows help() to work with update_wrapper
        out = lambda *args, **kwargs: self.fn(obj, *args, **kwargs)
        # update the docstring of the returned function
        update_wrapper(out, self.fn)
        return out


def if_delegate_has_method(delegate):
    """
        摘自sklearn中metaestimators.py if_delegate_has_method
        如果delegate有定义对应方法，才实际定义方法，否则被装饰的方法撤销
    """
    if isinstance(delegate, list):
        delegate = tuple(delegate)
    if not isinstance(delegate, tuple):
        delegate = (delegate,)

    return lambda fn: _IffHasAttrDescriptor(fn, delegate,
                                            attribute_name=fn.__name__)



# -*- encoding:utf-8 -*-
"""
    对各个依赖库不同版本，不同系统的规范进行统一以及问题修正模块
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import numbers
import sys

import matplotlib
import numpy as np
import pandas as pd
import scipy
import sklearn as skl

__author__ = '阿布'
__weixin__ = 'abu_quant'


def _parse_version(version_string):
    """
    根据库中的__version__字段，转换为tuple，eg. '1.11.3'->(1, 11, 3)
    :param version_string: __version__字符串对象
    :return: tuple 对象
    """
    version = []
    for x in version_string.split('.'):
        try:
            version.append(int(x))
        except ValueError:
            version.append(x)
    return tuple(version)

"""numpy 版本号tuple"""
np_version = _parse_version(np.__version__)
"""sklearn 版本号tuple"""
skl_version = _parse_version(skl.__version__)
"""pandas 版本号tuple"""
pd_version = _parse_version(pd.__version__)
"""scipy 版本号tuple"""
sp_version = _parse_version(scipy.__version__)
"""matplotlib 版本号tuple"""
mpl_version = _parse_version(matplotlib.__version__)


try:
    from inspect import signature, Parameter
except ImportError:
    try:
        from funcsigs import signature, Parameter
    except ImportError:
        from ..ExtBu.funcsigs import signature, Parameter

try:
    # noinspection PyCompatibility
    from concurrent.futures import ThreadPoolExecutor
except ImportError:
    from ..ExtBu.futures.thread import ThreadPoolExecutor

try:
    from ..ExtBu import six
except ImportError:
    import six as six

# try:
#     from six.moves import zip, xrange, range, reduce, map, filter
# except ImportError:
#     # noinspection PyUnresolvedReferences
#     from ..ExtBu.six.moves import zip, xrange, range, reduce, map, filter
# noinspection PyUnresolvedReferences
try:
    from ..ExtBu.six.moves import zip, xrange, range, reduce, map, filter
except ImportError:
    # noinspection PyUnresolvedReferences
    from six.moves import zip, xrange, range, reduce, map, filter

try:
    # noinspection all
    from six.moves import cPickle as pickle
except ImportError:
    # noinspection all
    from six.moves import cPickle as pickle

if six.PY3:
    # noinspection PyProtectedMember
    Unpickler = pickle._Unpickler
    # noinspection PyProtectedMember
    Pickler = pickle._Pickler
else:
    Unpickler = pickle.Unpickler
    Pickler = pickle.Pickler

if six.PY3:
    def as_bytes(s):
        if isinstance(s, bytes):
            return s
        return s.encode('latin1')
else:
    as_bytes = str

try:
    if six.PY3:
        from functools import lru_cache
    else:
        from functools32 import lru_cache
except ImportError:
    # noinspection PyUnusedLocal
    def lru_cache(maxsize=100):
        def decorate(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return decorate

try:
    from itertools import combinations_with_replacement
except ImportError:
    # Backport of itertools.combinations_with_replacement for Python 2.6,
    # from Python 3.4 documentation (http://tinyurl.com/comb-w-r), copyright
    # Python Software Foundation (https://docs.python.org/3/license.html)
    def combinations_with_replacement(iterable, r):
        # combinations_with_replacement('ABC', 2) --> AA AB AC BB BC CC
        pool = tuple(iterable)
        n = len(pool)
        if not n and r:
            return
        indices = [0] * r
        yield tuple(pool[i] for i in indices)
        while True:
            for i in reversed(range(r)):
                if indices[i] != n - 1:
                    break
            else:
                return
            indices[i:] = [indices[i] + 1] * (r - i)
            yield tuple(pool[i] for i in indices)

if sys.version_info < (2, 7, 0):
    # partial cannot be pickled in Python 2.6
    # http://bugs.python.org/issue1398
    # noinspection PyPep8Naming
    class partial(object):
        def __init__(self, func, *args, **keywords):
            functools.update_wrapper(self, func)
            self.func = func
            self.args = args
            self.keywords = keywords

        def __call__(self, *args, **keywords):
            args = self.args + args
            kwargs = self.keywords.copy()
            kwargs.update(keywords)
            return self.func(*args, **kwargs)
else:
    # noinspection PyUnresolvedReferences
    from functools import partial

"""
    matplotlib fixes
"""
# 先别加了，用的地方内部try吧，不然waring太多
# try:
#     # noinspection PyUnresolvedReferences, PyDeprecation
#     import matplotlib.finance as mpf
# except ImportError:
#     # 2.2 才会有
#     # noinspection PyUnresolvedReferences, PyDeprecation
#     import matplotlib.mpl_finance as mpf

"""
    urlencode
"""
if six.PY3:
    # noinspection PyUnresolvedReferences, PyCompatibility
    from urllib.parse import urlencode
else:
    # noinspection PyUnresolvedReferences
    from urllib import urlencode


"""
    sklearn fixes
"""


# noinspection PyProtectedMember,PyUnresolvedReferences
def check_random_state(seed):
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, (numbers.Integral, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                     ' instance' % seed)


if skl_version >= (0, 18, 0):
    mean_squared_error_scorer = 'neg_mean_squared_error'
    mean_absolute_error_scorer = 'neg_mean_absolute_error'
    median_absolute_error_scorer = 'neg_median_absolute_error'
    log_loss = 'neg_log_loss'

    try:
        from sklearn.model_selection import train_test_split
        from sklearn.model_selection import learning_curve
        from sklearn.model_selection import cross_val_score
        from sklearn.model_selection import GridSearchCV
        # noinspection PyPep8Naming
        from sklearn.mixture import GaussianMixture as GMM


        class KFold(object):
            """
                sklearn将KFold移动到了model_selection，而且改变了用法，暂时不需要
                这么复杂的功能，将sklearn中关键代码简单实现，不from sklearn.model_selection import KFold
            """
            def __init__(self, n, n_folds=3, shuffle=False, random_state=None):
                if abs(n - int(n)) >= np.finfo('f').eps:
                    raise ValueError("n must be an integer")
                self.n = int(n)

                if abs(n_folds - int(n_folds)) >= np.finfo('f').eps:
                    raise ValueError("n_folds must be an integer")
                self.n_folds = n_folds = int(n_folds)

                if n_folds <= 1:
                    raise ValueError(
                        "k-fold cross validation requires at least one"
                        " train / test split by setting n_folds=2 or more,"
                        " got n_folds={0}.".format(n_folds))
                if n_folds > self.n:
                    raise ValueError(
                        ("Cannot have number of folds n_folds={0} greater"
                         " than the number of samples: {1}.").format(n_folds, n))

                if not isinstance(shuffle, bool):
                    raise TypeError("shuffle must be True or False;"
                                    " got {0}".format(shuffle))
                self.shuffle = shuffle
                self.random_state = random_state

                self.idxs = np.arange(n)
                if shuffle:
                    rng = check_random_state(self.random_state)
                    rng.shuffle(self.idxs)

            def __iter__(self):
                ind = np.arange(self.n)
                for test_index in self._iter_test_masks():
                    train_index = np.logical_not(test_index)
                    train_index = ind[train_index]
                    test_index = ind[test_index]
                    yield train_index, test_index

            def _iter_test_masks(self):
                for test_index in self._iter_test_indices():
                    test_mask = self._empty_mask()
                    test_mask[test_index] = True
                    yield test_mask

            def _empty_mask(self):
                return np.zeros(self.n, dtype=np.bool)

            def _iter_test_indices(self):
                n = self.n
                n_folds = self.n_folds
                fold_sizes = (n // n_folds) * np.ones(n_folds, dtype=np.int)
                fold_sizes[:n % n_folds] += 1
                current = 0
                for fold_size in fold_sizes:
                    start, stop = current, current + fold_size
                    yield self.idxs[start:stop]
                    current = stop

            def __repr__(self):
                return '%s.%s(n=%i, n_folds=%i, shuffle=%s, random_state=%s)' % (
                    self.__class__.__module__,
                    self.__class__.__name__,
                    self.n,
                    self.n_folds,
                    self.shuffle,
                    self.random_state,
                )

            def __len__(self):
                return self.n_folds

    except ImportError:
        from sklearn.cross_validation import train_test_split
        from sklearn.cross_validation import KFold
        from sklearn.cross_validation import cross_val_score
        from sklearn.learning_curve import learning_curve
        from sklearn import cross_validation
        from sklearn.grid_search import GridSearchCV
        from sklearn.mixture import GMM
else:
    mean_squared_error_scorer = 'mean_squared_error'
    mean_absolute_error_scorer = 'mean_absolute_error'
    median_absolute_error_scorer = 'median_absolute_error'
    log_loss = 'log_loss'

    # noinspection PyUnresolvedReferences, PyDeprecation
    from sklearn.cross_validation import train_test_split
    # noinspection PyUnresolvedReferences, PyDeprecation
    from sklearn.cross_validation import KFold
    # noinspection PyUnresolvedReferences, PyDeprecation
    from sklearn.cross_validation import cross_val_score
    # noinspection PyUnresolvedReferences, PyDeprecation
    from sklearn.learning_curve import learning_curve
    # noinspection PyUnresolvedReferences, PyDeprecation
    from sklearn import cross_validation
    # noinspection PyUnresolvedReferences, PyDeprecation
    from sklearn.grid_search import GridSearchCV
    # noinspection PyUnresolvedReferences, PyDeprecation
    from sklearn.mixture import GMM

if np_version < (1, 8, 1):
    def array_equal(a1, a2):
        # copy-paste from numpy 1.8.1
        try:
            a1, a2 = np.asarray(a1), np.asarray(a2)
        except:
            return False
        if a1.shape != a2.shape:
            return False
        return bool(np.asarray(a1 == a2).all())
else:
    from numpy import array_equal

if sp_version < (0, 13, 0):
    def rankdata(a, method='average'):
        if method not in ('average', 'min', 'max', 'dense', 'ordinal'):
            raise ValueError('unknown method "{0}"'.format(method))

        arr = np.ravel(np.asarray(a))
        algo = 'mergesort' if method == 'ordinal' else 'quicksort'
        sorter = np.argsort(arr, kind=algo)

        inv = np.empty(sorter.size, dtype=np.intp)
        inv[sorter] = np.arange(sorter.size, dtype=np.intp)

        if method == 'ordinal':
            return inv + 1

        arr = arr[sorter]
        obs = np.r_[True, arr[1:] != arr[:-1]]
        dense = obs.cumsum()[inv]

        if method == 'dense':
            return dense

        # cumulative counts of each unique value
        # noinspection PyUnresolvedReferences
        count = np.r_[np.nonzero(obs)[0], len(obs)]

        if method == 'max':
            return count[dense]

        if method == 'min':
            return count[dense - 1] + 1

        # average method
        return .5 * (count[dense] + count[dense - 1] + 1)
else:
    # noinspection PyUnresolvedReferences
    from scipy.stats import rankdata



# -*- encoding:utf-8 -*-

"""
相对强弱指数（RSI）是通过比较一段时期内的平均收盘涨数和平均收盘跌数来分析市场买沽盘的意向和实力，
从而作出未来市场的走势

计算方法：

具体计算实现可阅读代码中_calc_rsi_from_pd()的实现
1. 根据收盘价格计算价格变动可以使用diff()也可以使用pct_change()
2. 分别筛选gain交易日的价格变动序列gain，和loss交易日的价格变动序列loss
3. 分别计算gain和loss的N日移动平均
4. rs = gain_mean / loss_mean
5. rsi = 100 - 100 / (1 + rs)

"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from .ABuNDBase import plot_from_order, g_calc_type, ECalcType
from ..UtilBu import ABuScalerUtil
from ..CoreBu.ABuPdHelper import pd_rolling_mean

__author__ = '阿布'
__weixin__ = 'abu_quant'

"""_calc_rsi_from_pd计算rs时使用gain，否则使用change"""
g_rsi_gain = True


# noinspection PyUnresolvedReferences
def _calc_rsi_from_ta(prices, time_period=14):
    """
    使用talib计算rsi, 即透传talib.RSI计算结果
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param time_period: rsi的N日参数, 默认14
    """

    import talib
    if isinstance(prices, pd.Series):
        prices = prices.values
    rsi = talib.RSI(prices, timeperiod=time_period)
    return rsi


# noinspection PyTypeChecker
def _calc_rsi_from_pd(prices, time_period=14):
    """
    通过rsi公式手动计算rsi
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param time_period: rsi的N日参数, 默认14
    """

    if not isinstance(prices, pd.Series):
        prices = pd.Series(prices)

    # 根据收盘价格计算价格变动可以使用diff()也可以使用pct_change()
    if g_rsi_gain:
        # 使用前后价格变动gain
        diff_price = prices.diff()
    else:
        # 使用前后价格变动change比例
        diff_price = prices.pct_change()
    diff_price[0] = 0

    # 分别筛选gain交易日的价格变动序列gain，和loss交易日的价格变动序列loss
    gain = np.where(diff_price > 0, diff_price, 0)
    loss = np.where(diff_price < 0, abs(diff_price), 0)
    # 分别计算gain和loss的N日移动平均
    gain_mean = pd_rolling_mean(gain, window=time_period)
    loss_mean = pd_rolling_mean(loss, window=time_period)
    # rsi = 100 - 100 / (1 +  gain_mean / loss_mean)
    rs = gain_mean / loss_mean
    rsi = 100 - 100 / (1 + rs)
    return rsi


"""通过在ABuNDBase中尝试import talib来统一确定指标计算方式, 外部计算只应该使用calc_rsi"""
calc_rsi = _calc_rsi_from_pd if g_calc_type == ECalcType.E_FROM_PD else _calc_rsi_from_ta


def plot_rsi_from_order(order, date_ext=120, **kwargs):
    """
    封装ABuNDBase中的plot_from_order与模块中绘制技术指标的函数，完成技术指标可视化及标注买入卖出点位
    :param order: AbuOrder对象转换的pd.DataFrame对象or pd.Series对象
    :param date_ext: int对象 eg. 如交易在2015-06-01执行，如date_ext＝120，择start向前推120天，end向后推120天
    :param kwargs: 绘制技术指标需要的其它关键字参数，time_period, 最终透传给plot_rsi
    """
    return plot_from_order(plot_rsi_from_klpd, order, date_ext, **kwargs)


def plot_rsi_from_klpd(kl_pd, with_points=None, with_points_ext=None, **kwargs):
    """
    封装plot_rsi，绘制收盘价格，rsi曲线
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param with_points: 这里的常规用途是传入买入order, with_points=buy_index=pd.to_datetime(orders['buy_date']))
    :param with_points_ext: 这里的常规用途是传入卖出order, with_points_ext=sell_index=pd.to_datetime(orders['sell_date']))
    :param kwargs: 绘制技术指标需要的其它关键字参数，time_period, 最终透传给plot_rsi
    """

    plot_rsi(kl_pd.close, kl_pd.index, with_points=with_points, with_points_ext=with_points_ext,
             **kwargs)


def plot_rsi(prices, kl_index, with_points=None, with_points_ext=None, with_price=True, time_period=14):
    """
    绘制收盘价格，以及对应的macd曲线，如果有with_points点位标注，使用竖线标注
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param kl_index: pd.Index时间序列
    :param with_points: 这里的常规用途是传入买入order, with_points=buy_index=pd.to_datetime(orders['buy_date']))
    :param with_points_ext: 这里的常规用途是传入卖出order, with_points_ext=sell_index=pd.to_datetime(orders['sell_date']))
    :param with_price: 将价格一起绘制，但两个曲线进行缩放到一个数值级别
    :param time_period: rsi的N日参数, 默认14
    :return:
    """
    # noinspection PyTypeChecker
    rsi = np.array(calc_rsi(prices, time_period))

    plt.figure(figsize=[16, 8])
    plt.axes([0.025, 0.025, 0.95, 0.95])

    if with_price:
        # 绘制在一个画布上, 将两个曲线进行缩放到一个数值级别
        matrix = ABuScalerUtil.scaler_matrix([rsi, prices])
        rsi, prices = matrix[matrix.columns[0]], matrix[matrix.columns[1]]
        plt.plot(kl_index, prices, label='close price')

    plt.plot(kl_index, rsi, label='rsi')

    # with_points和with_points_ext的点位使用竖线标注
    if with_points is not None:
        plt.axvline(with_points, color='green', linestyle='--')

    if with_points_ext is not None:
        plt.axvline(with_points_ext, color='red')

    plt.grid(True)
    plt.legend(loc='best')
    plt.show()



# -*- encoding:utf-8 -*-
"""
    买入卖出因子与ump进行组织管理通信模块
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from ..UtilBu.ABuLazyUtil import LazyFunc
from ..UtilBu.ABuFileUtil import file_exist
from ..UtilBu.ABuDelegateUtil import first_delegate_has_method, replace_word_delegate_has_method
from ..UmpBu.ABuUmpEdgeBase import EEdgeType
from ..UmpBu.ABuUmpEdgeDeg import AbuUmpEdgeDeg
from ..UmpBu.ABuUmpEdgeFull import AbuUmpEdgeFull
from ..UmpBu.ABuUmpEdgePrice import AbuUmpEdgePrice
from ..UmpBu.ABuUmpEdgeWave import AbuUmpEdgeWave
from ..UmpBu.ABuUmpMainDeg import AbuUmpMainDeg
from ..UmpBu.ABuUmpMainJump import AbuUmpMainJump
from ..UmpBu.ABuUmpMainPrice import AbuUmpMainPrice
from ..UmpBu.ABuUmpMainWave import AbuUmpMainWave
from ..UmpBu.ABuUmpMainBase import AbuUmpMainBase
from ..CoreBu import ABuEnv
# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import filter, six

__author__ = '阿布'
__weixin__ = 'abu_quant'

"""外部用户使用append_user_ump添加到外部ump list容器中"""
_g_extend_ump_list = list()
"""是否启用外部用户使用append_user_ump添加的ump对交易进行拦截决策"""
g_enable_user_ump = False


def append_user_ump(ump, check=True):
    """
    外部用户设置扩展ump接口, 参数ump可以是ump class类型，也可以是实例化后的ump object
    :param ump: 可以是ump class类型，也可以是实例化后的ump object
    :param check: 是否检测ump是否以及训练好，且生成了cache文件
    """
    if check:
        # 检测ump训练后的本地物理文件是否存在
        if isinstance(ump, six.class_types):
            ump_cache_path = ump(predict=True).dump_file_fn()
        else:
            ump_cache_path = ump.dump_file_fn()
        if not file_exist(ump_cache_path):
            # 如果用户添加的ump没有经过训练集训练，提升用户进行训练
            raise RuntimeError('you must first fit orders, {} is not exist!!'.format(ump_cache_path))
    _g_extend_ump_list.append(ump)


def clear_user_ump():
    """
    将外部用户设置的ump队列清空
    :return:
    """
    global _g_extend_ump_list
    _g_extend_ump_list = list()


class AbuUmpManager(object):
    """ump组织管理类"""

    def __init__(self, factor):
        """
        在AbuFactorBuyBase __init__和AbuFactorSellBase __init__函数中构造AbuUmpManager，
        外部用户不应有构造

        :param factor: 买入或者卖出因子对象，AbuFactorBuyBase or AbuFactorSellBase类对象实例
        """

        # 在买入或卖出因子__init__ 中self.ump_manger = AbuUmpManager(self)，即因子和AbuUmpManager互持op
        self.factor = factor
        from ..FactorBuyBu import AbuFactorBuyBase
        self.is_buy_factor = isinstance(self.factor, AbuFactorBuyBase)

        self.extend_ump_list = list()
        if g_enable_user_ump and len(_g_extend_ump_list) > 0:
            # 筛选与因子对应的ump，即买入因子AbuFactorBuyBase对象只筛选买入BuyUmpMixin类型的ump
            filter_ump = list(filter(lambda ump: self.is_buy_factor == ump.is_buy_ump(), _g_extend_ump_list))
            self.extend_ump_list.extend(filter_ump)
        self._fix_ump_env()

    def __str__(self):
        """打印对象显示：class name, factor, extend_ump_list"""
        return '{}: factor={}, self.extend_ump_list={}'.format(self.__class__.__name__, self.factor,
                                                               self.extend_ump_list)

    __repr__ = __str__

    @classmethod
    def _fix_ump_env(cls):
        """
        只为修复在ABuEnv中对ump的设置和manger中的ump设置不同步问题，即ABuEnv.g_enable_ml_feature
        设置不正确，将env中ump的设置迁移到本模块中后便不需要fix了，外部user不应主动使用，只在manger构造使用
        """
        if ABuEnv.g_enable_ump_edge_deg_block or ABuEnv.g_enable_ump_edge_price_block or \
                ABuEnv.g_enable_ump_edge_wave_block or ABuEnv.g_enable_ump_edge_full_block:
            ABuEnv.g_enable_ml_feature = True
        if ABuEnv.g_enable_ump_main_deg_block or ABuEnv.g_enable_ump_main_jump_block or \
                ABuEnv.g_enable_ump_main_price_block or ABuEnv.g_enable_ump_main_wave_block:
            ABuEnv.g_enable_ml_feature = True
        # TODO 将env中ump的设置迁移到本模块

        if g_enable_user_ump:
            ABuEnv.g_enable_ml_feature = True

    @LazyFunc
    def ump_main_deg(self):
        """
            根据ABuEnv.g_enable_ump_main_deg_block设置决定是否构造AbuUmpMainDeg，
            使用LazyFunc装饰器，构造后即使用缓存AbuUmpMainDeg(predict=True)实例
        """
        return AbuUmpMainDeg(predict=True)

    @LazyFunc
    def ump_main_jump(self):
        """
            根据ABuEnv.g_enable_ump_main_jump_block设置决定是否构造AbuUmpMainJump，
            使用LazyFunc装饰器，构造后即使用缓存AbuUmpMainJump(predict=True)实例
        """
        return AbuUmpMainJump(predict=True)

    @LazyFunc
    def ump_main_price(self):
        """
            根据ABuEnv.g_enable_ump_main_price_block设置决定是否构造AbuUmpMainPrice，
            使用LazyFunc装饰器，构造后即使用缓存AbuUmpMainPrice(predict=True)实例
        """
        return AbuUmpMainPrice(predict=True)

    @LazyFunc
    def ump_main_wave(self):
        """
            根据ABuEnv.g_enable_ump_main_wave_block设置决定是否构造AbuUmpMainWave，
            使用LazyFunc装饰器，构造后即使用缓存AbuUmpMainWave(predict=True)实例
        """
        return AbuUmpMainWave(predict=True)

    @LazyFunc
    def ump_edge_deg(self):
        """
            根据ABuEnv.g_enable_ump_edge_deg_block设置决定是否构造AbuUmpEdgeDeg，
            使用LazyFunc装饰器，构造后即使用缓存AbuUmpEdgeDeg(predict=True)实例
        """
        return AbuUmpEdgeDeg(predict=True)

    @LazyFunc
    def ump_edge_price(self):
        """
            根据ABuEnv.g_enable_ump_edge_price_block设置决定是否构造AbuUmpEdgePrice，
            使用LazyFunc装饰器，构造后即使用缓存AbuUmpEdgePrice(predict=True)实例
        """
        return AbuUmpEdgePrice(predict=True)

    @LazyFunc
    def ump_edge_wave(self):
        """
            根据ABuEnv.g_enable_ump_edge_wave_block设置决定是否构造AbuUmpEdgeWave，
            使用LazyFunc装饰器，构造后即使用缓存AbuUmpEdgeWave(predict=True)实例
        """
        return AbuUmpEdgeWave(predict=True)

    @LazyFunc
    def ump_edge_full(self):
        """
            根据ABuEnv.g_enable_ump_edge_full_block设置决定是否构造AbuUmpEdgeFull，
            使用LazyFunc装饰器，构造后即使用缓存AbuUmpEdgeFull(predict=True)实例
        """
        return AbuUmpEdgeFull(predict=True)

    # noinspection PyMethodMayBeStatic
    def _default_main_hit_cnt(self):
        # noinspection PyProtectedMember
        default_hit_cnt = 1 if ABuEnv._g_enable_example_env_ipython else 2
        return default_hit_cnt

    @first_delegate_has_method(delegate='factor')
    def ump_main_deg_hit_cnt(self):
        """
        角度主裁使用predict_kwargs的参数need_hit_cnt值，即:
                predict_kwargs(need_hit_cnt=self.ump_main_deg_hit_cnt(), **ml_feature_dict)

        被装饰器first_delegate_has_method(delegate='factor')装饰，当被委托的因子，即self.factor中
        有对应实现的ump_main_deg_hit_cnt方法时，返回self.factor.ump_main_deg_hit_cnt()的返回值
        :return: int
        """
        return self._default_main_hit_cnt()

    @first_delegate_has_method(delegate='factor')
    def ump_main_jump_hit_cnt(self):
        """
        跳空主裁使用predict_kwargs的参数need_hit_cnt值，即:
                predict_kwargs(need_hit_cnt=self.ump_main_jump_hit_cnt, **ml_feature_dict)

        被装饰器first_delegate_has_method(delegate='factor')装饰，当被委托的因子，即self.factor中
        有对应实现的ump_main_jump_hit_cnt方法时，返回self.factor.ump_main_jump_hit_cnt()的返回值
        :return: int
        """
        return self._default_main_hit_cnt()

    @first_delegate_has_method(delegate='factor')
    def ump_main_price_hit_cnt(self):
        """
        价格主裁使用predict_kwargs的参数need_hit_cnt值，即:
                predict_kwargs(need_hit_cnt=self.ump_main_price_hit_cnt, **ml_feature_dict)

        被装饰器first_delegate_has_method(delegate='factor')装饰，当被委托的因子，即self.factor中
        有对应实现的ump_main_price_hit_cnt方法时，返回self.factor.ump_main_price_hit_cnt()的返回值
        :return: int
        """
        return self._default_main_hit_cnt()

    @first_delegate_has_method(delegate='factor')
    def ump_main_wave_hit_cnt(self):
        """
        价格波动主裁使用predict_kwargs的参数need_hit_cnt值，即:
                predict_kwargs(need_hit_cnt=self.ump_main_wave_hit_cnt, **ml_feature_dict)

        被装饰器first_delegate_has_method(delegate='factor')装饰，当被委托的因子，即self.factor中
        有对应实现的ump_main_wave_hit_cnt方法时，返回self.factor.ump_main_wave_hit_cnt()的返回值
        :return: int
        """
        return self._default_main_hit_cnt()

    # noinspection PyMethodMayBeStatic
    def ump_main_user_hit_cnt(self):
        """
        用户自定义的主裁ump类的使用predict_kwargs的参数need_hit_cnt值，
        用户在因子中可通过实现特点的方法名称来替换ump_main_user_hit_cnt值，

        使用时使用replace_word_delegate_has_method检测self.factor中有没有对应的方法，
        eg:
            replace_hit_cnt = replace_word_delegate_has_method(delegate='factor', key_word='user',
                                                                       replace_word=class_unique_id)
            hit_cnt = replace_hit_cnt(self.ump_main_user_hit_cnt)()

            即如果用户编写的主裁ump中class_unique_id方法返回'extend_main_test'
                @classmethod
                def class_unique_id(cls):
                    return 'extend_main_test'
            则在因子中对应自定义hit_cnt的方法名称应为：
                def ump_main_extend_main_test_hit_cnt(self)
                    return 1
            更多具体实现阅extend_ump_block以及replace_word_delegate_has_method方法的实现

        :return: int
        """
        return self._default_main_hit_cnt()

    def ump_block(self, ml_feature_dict):
        """
        在买入或者卖出因子中make_ump_block_decision方法中使用，决策特定交易是否被拦截，
        ump_block中首先使用内置ump进行拦截决策，如果不被拦截，使用外部定义的ump进行拦截决策

        :param ml_feature_dict: 交易所形成的特征字典
                eg: ml_feature_dict
                    {'buy_deg_ang42': -0.45400000000000001, 'buy_deg_ang252': 5.532,
                    'buy_deg_ang60': 2.1419999999999999, 'buy_deg_ang21': 0.93100000000000005,
                    'buy_price_rank120': 1.0, 'buy_price_rank90': 1.0, 'buy_price_rank60': 1.0,
                    'buy_price_rank252': 1.0, 'buy_wave_score1': 1.2470000000000001, 'buy_wave_score2': 1.286,
                    'buy_wave_score3': 1.2849999999999999, 'buy_atr_std': 0.19400000000000001,
                    'buy_jump_down_power': -13.57, 'buy_diff_down_days': 136, 'buy_jump_up_power': 1.038,
                    'buy_diff_up_days': 2}
        :return: bool, 对ml_feature_dict所描述的交易特征是否进行拦截
        """
        # 内置ump进行拦截决策
        if self.builtin_ump_block(ml_feature_dict):
            return True

        # 外部定义的ump进行拦截决策
        if self.extend_ump_block(ml_feature_dict):
            return True
        return False

    def extend_ump_block(self, ml_feature_dict):
        """
        外部用户设置的ump进行拦截决策，迭代self.extend_ump_list中外部设置的ump，
        由于对外添加ump的接口append_user_ump中参数ump可以是ump class类型，
        也可以是实例化后的ump object，所以需要把class类型的ump进行实例构造，且将
        实例的ump对象缓存在类变量中（通过class_unique_id为类变量构造唯一名称），
        ump对象构造好后根据主裁还是边裁选择决策方法：
                主裁使用：predict_kwargs(need_hit_cnt=need_hit_cnt, **ml_feature_dict)
                边裁使用：predict(**ml_feature_dict) == EEdgeType.E_EEdge_TOP_LOSS
        对交易进行拦截决策

        :param ml_feature_dict: 交易所形成的特征字典
        eg: ml_feature_dict
            {'buy_deg_ang42': -0.45400000000000001, 'buy_deg_ang252': 5.532,
            'buy_deg_ang60': 2.1419999999999999, 'buy_deg_ang21': 0.93100000000000005,
            'buy_price_rank120': 1.0, 'buy_price_rank90': 1.0, 'buy_price_rank60': 1.0,
            'buy_price_rank252': 1.0, 'buy_wave_score1': 1.2470000000000001, 'buy_wave_score2': 1.286,
            'buy_wave_score3': 1.2849999999999999, 'buy_atr_std': 0.19400000000000001,
            'buy_jump_down_power': -13.57, 'buy_diff_down_days': 136, 'buy_jump_up_power': 1.038,
            'buy_diff_up_days': 2}
        :return: bool, 对ml_feature_dict所描述的交易特征是否进行拦截
        """
        for extend_ump in self.extend_ump_list:
            class_unique_id = extend_ump.class_unique_id()
            # 由于对外添加ump的接口append_user_ump中参数ump可以是ump class类型，也可以是实例化后的ump object
            if isinstance(extend_ump, six.class_types):
                # 把class类型的ump进行实例构造

                is_main_ump = issubclass(extend_ump, AbuUmpMainBase)
                main_ump_key = 'main' if is_main_ump else 'edge'

                # 通过class_unique_id和issubclass(extend_ump, AbuUmpMainBase)为类变量构造唯一名称
                extend_ump_attr_str = 'ump_{}_{}'.format(main_ump_key, class_unique_id)
                if hasattr(self, extend_ump_attr_str):
                    # 将类变量中的实例代替类
                    extend_ump = getattr(self, extend_ump_attr_str)
                else:
                    # 内置ump通过LazyFunc进行效率提升，外部设置的ump通过手动setattr，将实例的ump对象缓存在类变量中
                    extend_ump_obj = extend_ump(predict=True)
                    setattr(self, extend_ump_attr_str, extend_ump_obj)
                    # 将实例化后的实例代替类
                    extend_ump = extend_ump_obj

            is_main_ump = isinstance(extend_ump, AbuUmpMainBase)
            if is_main_ump:
                # replace_word_delegate_has_method不做装饰器修饰ump_main_user_hit_cnt，因为需要动态获取replace_word
                try:
                    replace_hit_cnt = replace_word_delegate_has_method(delegate='factor', key_word='user',
                                                                       replace_word=class_unique_id)
                    hit_cnt = replace_hit_cnt(self.ump_main_user_hit_cnt)()
                except:
                    # 忽略用户自定义factor中关于hit_cnt的任何错误
                    hit_cnt = self.ump_main_user_hit_cnt()
                if extend_ump.predict_kwargs(need_hit_cnt=hit_cnt, **ml_feature_dict):
                    return True
            else:
                if extend_ump.predict(**ml_feature_dict) == EEdgeType.E_EEdge_TOP_LOSS:
                    return True
        return False

    def builtin_ump_block(self, ml_feature_dict):
        """
        内置ump进行拦截决策，通过ABuEnv中的拦截设置以及因子的买入卖出类型是否和ump类型匹配，
        来决定是否使用特定的ump进行拦截决策，如需要决策：

                主裁使用：predict_kwargs(need_hit_cnt=need_hit_cnt, **ml_feature_dict)
                边裁使用：predict(**ml_feature_dict) == EEdgeType.E_EEdge_TOP_LOSS
        对交易进行拦截决策

                :param ml_feature_dict: 交易所形成的特征字典
                eg: ml_feature_dict
                    {'buy_deg_ang42': -0.45400000000000001, 'buy_deg_ang252': 5.532,
                    'buy_deg_ang60': 2.1419999999999999, 'buy_deg_ang21': 0.93100000000000005,
                    'buy_price_rank120': 1.0, 'buy_price_rank90': 1.0, 'buy_price_rank60': 1.0,
                    'buy_price_rank252': 1.0, 'buy_wave_score1': 1.2470000000000001, 'buy_wave_score2': 1.286,
                    'buy_wave_score3': 1.2849999999999999, 'buy_atr_std': 0.19400000000000001,
                    'buy_jump_down_power': -13.57, 'buy_diff_down_days': 136, 'buy_jump_up_power': 1.038,
                    'buy_diff_up_days': 2}

        :return: bool, 对ml_feature_dict所描述的交易特征是否进行拦截
        """

        """内置主裁开始"""
        if ABuEnv.g_enable_ump_main_deg_block and self.is_buy_factor == self.ump_main_deg.is_buy_ump() \
                and self.ump_main_deg.predict_kwargs(need_hit_cnt=self.ump_main_deg_hit_cnt(), **ml_feature_dict):
            return True

        if ABuEnv.g_enable_ump_main_jump_block and self.is_buy_factor == self.ump_main_jump.is_buy_ump() \
                and self.ump_main_jump.predict_kwargs(need_hit_cnt=self.ump_main_jump_hit_cnt(), **ml_feature_dict):
            return True

        if ABuEnv.g_enable_ump_main_price_block and self.is_buy_factor == self.ump_main_price.is_buy_ump() \
                and self.ump_main_price.predict_kwargs(need_hit_cnt=self.ump_main_price_hit_cnt(), **ml_feature_dict):
            return True

        if ABuEnv.g_enable_ump_main_wave_block and self.is_buy_factor == self.ump_main_wave.is_buy_ump() \
                and self.ump_main_wave.predict_kwargs(need_hit_cnt=self.ump_main_wave_hit_cnt(), **ml_feature_dict):
            return True

        """内置边裁开始"""
        if ABuEnv.g_enable_ump_edge_deg_block and self.is_buy_factor == self.ump_edge_deg.is_buy_ump() \
                and self.ump_edge_deg.predict(**ml_feature_dict) == EEdgeType.E_EEdge_TOP_LOSS:
            return True

        if ABuEnv.g_enable_ump_edge_price_block and self.is_buy_factor == self.ump_edge_price.is_buy_ump() \
                and self.ump_edge_price.predict(**ml_feature_dict) == EEdgeType.E_EEdge_TOP_LOSS:
            return True

        if ABuEnv.g_enable_ump_edge_wave_block and self.is_buy_factor == self.ump_edge_wave.is_buy_ump() \
                and self.ump_edge_wave.predict(**ml_feature_dict) == EEdgeType.E_EEdge_TOP_LOSS:
            return True

        if ABuEnv.g_enable_ump_edge_full_block and self.is_buy_factor == self.ump_edge_full.is_buy_ump() \
                and self.ump_edge_full.predict(**ml_feature_dict) == EEdgeType.E_EEdge_TOP_LOSS:
            return True

        return False



# -*- encoding:utf-8 -*-
"""
    移动平均线，Moving Average，简称MA，原本的意思是移动平均，由于我们将其制作成线形，所以一般称之为移动平均线，简称均线。
    它是将某一段时间的收盘价之和除以该周期。 比如日线MA5指5天内的收盘价除以5 。
    移动平均线是由著名的美国投资专家Joseph E.Granville（葛兰碧，又译为格兰威尔）于20世纪中期提出来的。
    均线理论是当今应用最普遍的技术指标之一，它帮助交易者确认现有趋势、判断将出现的趋势、发现过度延生即将反转的趋势
"""

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

from collections import Iterable

import matplotlib.pyplot as plt
import pandas as pd
from enum import Enum

from .ABuNDBase import plot_from_order, g_calc_type, ECalcType
from ..CoreBu.ABuPdHelper import pd_rolling_mean, pd_ewm_mean
from ..CoreBu.ABuFixes import six
from ..UtilBu.ABuDTUtil import catch_error

__author__ = '阿布'
__weixin__ = 'abu_quant'


class EMACalcType(Enum):
    """计算移动移动平均使用的方法"""
    """简单移动平均线"""
    E_MA_MA = 0
    """加权移动平均线"""
    E_MA_EMA = 1


# noinspection PyUnresolvedReferences
def _calc_ma_from_ta(prices, time_period=10, from_calc=EMACalcType.E_MA_MA):
    """
    使用talib计算ma，即透传talib.MA or talib.EMA计算结果
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param time_period: 移动平均的N值，int
    :param from_calc: EMACalcType enum对象，移动移动平均使用的方法
    """

    import talib
    if isinstance(prices, pd.Series):
        prices = prices.values

    if from_calc == EMACalcType.E_MA_MA:
        ma = talib.MA(prices, timeperiod=time_period)
    else:
        ma = talib.EMA(prices, timeperiod=time_period)
    return ma


def _calc_ma_from_pd(prices, time_period=10, from_calc=EMACalcType.E_MA_MA):
    """
    通过pandas计算ma或者ema
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param time_period: 移动平均的N值，int
    :param from_calc: EMACalcType enum对象，移动移动平均使用的方法
    """

    if isinstance(prices, pd.Series):
        prices = prices.values

    if from_calc == EMACalcType.E_MA_MA:
        ma = pd_rolling_mean(prices, window=time_period, min_periods=time_period)
    else:
        ma = pd_ewm_mean(prices, span=time_period, min_periods=time_period)
    return ma


def calc_ma_from_prices(prices, time_period=10, min_periods=None, from_calc=EMACalcType.E_MA_MA):
    """
    通过pandas计算ma或者ema, 添加min_periods参数
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param time_period: 移动平均的N值，int
    :param min_periods: int，默认None则使用time_period
    :param from_calc: EMACalcType enum对象，移动移动平均使用的方法
    """

    if isinstance(prices, pd.Series):
        prices = prices.values

    min_periods = time_period if min_periods is None else min_periods
    if from_calc == EMACalcType.E_MA_MA:
        ma = pd_rolling_mean(prices, window=time_period, min_periods=min_periods)
    else:
        ma = pd_ewm_mean(prices, span=time_period, min_periods=min_periods)
    return ma


"""通过在ABuNDBase中尝试import talib来统一确定指标计算方式, 外部计算只应该使用calc_ma"""
calc_ma = _calc_ma_from_pd if g_calc_type == ECalcType.E_FROM_PD else _calc_ma_from_ta


def plot_ma_from_order(order, date_ext=120, **kwargs):
    """
    封装ABuNDBase中的plot_from_order与模块中绘制技术指标的函数，完成技术指标可视化及标注买入卖出点位
    :param order: AbuOrder对象转换的pd.DataFrame对象or pd.Series对象
    :param date_ext: int对象 eg. 如交易在2015-06-01执行，如date_ext＝120，择start向前推120天，end向后推120天
    :param kwargs: 绘制技术指标需要的其它关键字参数，time_period，from_calc, with_price，最终透传给plot_ma
    """
    return plot_from_order(plot_ma_from_klpd, order, date_ext, **kwargs)


def plot_ma_from_klpd(kl_pd, with_points=None, with_points_ext=None, **kwargs):
    """
    封装plot_ma，绘制收盘价格，以及多条移动均线
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param with_points: 这里的常规用途是传入买入order, with_points=buy_index=pd.to_datetime(orders['buy_date']))
    :param with_points_ext: 这里的常规用途是传入卖出order, with_points_ext=sell_index=pd.to_datetime(orders['sell_date']))
    :param kwargs: 绘制技术指标需要的其它关键字参数，time_period，from_calc, with_price，最终透传给plot_ma
    """

    # 如果外部不设置均线，这里pop的default为[30, 60, 90]
    time_period = kwargs.pop('time_period', [30, 60, 90])
    plot_ma(kl_pd.close, kl_pd.index, time_period, with_points=with_points,
            with_points_ext=with_points_ext)


def plot_ma(prices, kl_index, time_period, from_calc=EMACalcType.E_MA_MA,
            with_points=None, with_points_ext=None, with_price=True):
    """
    一个画布上，绘制价格曲线以及多条移动平均线，如果有with_points点位标注，使用竖线标注
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param kl_index: pd.Index时间序列
    :param time_period: 注意是Iterable类型，需要可迭代对象，即使一个元素也要使用如[10]包裹
    :param from_calc: EMACalcType enum对象，默认使用简单移动平均线
    :param with_points: 这里的常规用途是传入买入order, with_points=buy_index=pd.to_datetime(orders['buy_date']))
    :param with_points_ext: 这里的常规用途是传入卖出order, with_points_ext=sell_index=pd.to_datetime(orders['sell_date']))
    :param with_price:  将价格一起绘制
    :return:
    """
    # TODO Iterable和six.string_types的判断抽出来放在一个模块，做为Iterable的判断来使用
    if not isinstance(time_period, Iterable) or isinstance(time_period, six.string_types):
        raise TypeError('MA CALC time_period MUST PASS Iterable!!!')

    calc_type_func = calc_ma
    # 迭代计算多条移动均线，使用from_calc使用的方法计算
    ma_array = [calc_type_func(prices, period, from_calc) for period in time_period]

    plt.figure(figsize=[14, 7])

    for ind, ma in enumerate(ma_array):
        # ind的作用是索引在原始time_period中对应label需要的名称
        # noinspection PyUnresolvedReferences
        plt.plot(kl_index, ma, label='ma {}'.format(time_period[ind]))

    if with_price:
        plt.plot(kl_index, prices, label='prices')

    @catch_error(return_val=None, log=False)
    def plot_with_point(points, co, cc):
        """
        点位使用圆点＋竖线进行标注
        :param points: 点位坐标序列
        :param co: 点颜色 eg. 'go' 'ro'
        :param cc: markeredgecolor和竖线axvline颜色 eg. 'green' 'red'
        """
        v_index_num = kl_index.tolist().index(points)
        # 如果有ma线，y点做目标画在第一根ma线上否则画在价格上面
        y_array = ma_array[0] if len(ma_array) > 0 else prices
        plt.plot(points, y_array[v_index_num], co, markersize=12, markeredgewidth=3.0,
                 markerfacecolor='None', markeredgecolor=cc)
        plt.axvline(points, color=cc)

    # with_points和with_points_ext的点位使用竖线标注
    if with_points is not None:
        # plt.axvline(with_points, color='green', linestyle='--')
        plot_with_point(with_points, 'go', 'green')

    if with_points_ext is not None:
        # plt.axvline(with_points_ext, color='red')
        plot_with_point(with_points_ext, 'ro', 'red')

    plt.grid(True)
    plt.legend(loc='best')
    plt.show()



from __future__ import absolute_import

from .ABuSlippageBuyBase import AbuSlippageBuyBase
from .ABuSlippageBuyMean import AbuSlippageBuyMean
from .ABuSlippageSellBase import AbuSlippageSellBase
from .ABuSlippageSellMean import AbuSlippageSellMean

from . import ABuSlippage as slippage

__all__ = [
    'AbuSlippageBuyBase',
    'AbuSlippageBuyMean',
    'AbuSlippageSellBase',
    'AbuSlippageSellMean',

    'slippage']



# -*- encoding:utf-8 -*-
"""量化技术分析工具图形可视化"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import logging
from contextlib import contextmanager

from IPython.display import display
import ipywidgets as widgets

from ..TLineBu.ABuTL import EShiftDistanceHow, ESkeletonHow
from ..TLineBu.ABuTLExecute import calc_pair_speed
from ..TLineBu.ABuTLJump import calc_jump, calc_jump_line, calc_jump_line_weight
from ..TLineBu.ABuTLGolden import calc_golden
from ..UtilBu import ABuProgress
from ..WidgetBu.ABuWGToolBase import WidgetToolBase, single_fetch_symbol_analyse

__author__ = '阿布'
__weixin__ = 'abu_quant'


# noinspection PyUnusedLocal
class WidgetTLTool(WidgetToolBase):
    """技术分析界面"""

    @contextmanager
    def _init_tip_label_with_step_x(self, callback_analyse, analyse_name, with_step_x=True):
        """step_x需要的地方比较多，统一构建，外部接收赋予名字"""
        if not callable(callback_analyse):
            raise TabError('callback_analyse must callable!')

        tip_label = widgets.Label(self.map_tip_target_label(n_target=1), layout=self.label_layout)
        widget_list = [tip_label]
        step_x = None
        if with_step_x:
            step_x_label = widgets.Label(u'时间步长控制参数step_x，默认1.0',
                                         layout=self.label_layout)
            step_x = widgets.FloatSlider(
                value=1.0,
                min=0.1,
                max=2.6,
                step=0.1,
                description=u'步长',
                disabled=False,
                orientation='horizontal',
                readout=True,
                readout_format='.1f',
            )
            # 返回给需要的ui，命名独有的step_x
        yield widget_list, step_x

        if with_step_x:
            # noinspection PyUnboundLocalVariable
            step_x_box = widgets.VBox([step_x_label, step_x])
            # noinspection PyTypeChecker
            widget_list.append(step_x_box)

        analyse_bt = widgets.Button(description=analyse_name, layout=widgets.Layout(width='98%'),
                                    button_style='info')
        analyse_bt.on_click(callback_analyse)
        widget_list.append(analyse_bt)

    def init_rs_ui(self):
        """阻力支撑分析ui"""

        with self._init_tip_label_with_step_x(
                self._rs_line_analyse, u'支撑阻力线分析', with_step_x=False) as (widget_list, _):
            self.rs_mode = widgets.RadioButtons(
                options={u'只分析支撑线': 0, u'只分析阻力线': 1, u'支撑线和阻力线': 2},
                value=0,
                description=u'分析模式:',
                disabled=False
            )
            widget_list.append(self.rs_mode)
            self.only_last = widgets.RadioButtons(
                options={u'最近的阻力线和支撑线': True, u'所有的阻力线和支撑线': False},
                value=True,
                description=u'最近的阻力线和支撑线',
                disabled=False
            )
            widget_list.append(self.only_last)
        return widgets.VBox(widget_list,  # border='solid 1px',
                            layout=self.tool_layout)

    def init_jump_ui(self):
        """跳空分析ui"""

        with self._init_tip_label_with_step_x(
                self._jump_line_analyse, u'跳空技术分析', with_step_x=False) as (widget_list, _):
            self.jump_mode = widgets.RadioButtons(
                options={u'跳空统计分析': 0, u'跳空缺口筛选': 1, u'缺口时间加权筛选': 2},
                value=0,
                description=u'分析模式:',
                disabled=False
            )
            widget_list.append(self.jump_mode)

            power_threshold_label = widgets.Label(u'缺口能量阀值，默认2.0(只对缺口筛选生效)',
                                                  layout=self.label_layout)
            self.power_threshold = widgets.FloatSlider(
                value=2.0,
                min=1.5,
                max=3.5,
                step=0.1,
                description=u'能量',
                disabled=False,
                orientation='horizontal',
                readout=True,
                readout_format='.1f',
            )
            power_threshold_box = widgets.VBox([power_threshold_label, self.power_threshold])
            widget_list.append(power_threshold_box)

            jump_diff_factor_label = widgets.Label(u'设置调节跳空阀值的大小',
                                                   layout=self.label_layout)
            self.jump_diff_factor = widgets.FloatSlider(
                value=1.0,
                min=0.1,
                max=5.0,
                step=0.1,
                description=u'阀值',
                disabled=False,
                orientation='horizontal',
                readout=True,
                readout_format='.1f',
            )
            jump_diff_factor_box = widgets.VBox([jump_diff_factor_label, self.jump_diff_factor])
            widget_list.append(jump_diff_factor_box)

        return widgets.VBox(widget_list,
                            # border='solid 1px',
                            layout=self.tool_layout)

    def init_shift_distance_ui(self):
        """位移路程比ui"""

        with self._init_tip_label_with_step_x(
                self._shift_distance_analyse, u'位移路程分析') as (widget_list, step_x):
            self.shift_distance_step_x = step_x
            self.shift_distance_mode = widgets.RadioButtons(
                options={u'序列最后的元素做为路程基础': 0, u'极限值做为路程的计算基础': 1,
                         u'序列sum+极值做为路程计算基础': 2},
                value=0,
                description=u'路程模式:',
                disabled=False
            )
            widget_list.append(self.shift_distance_mode)
        return widgets.VBox(widget_list,
                            # border='solid 1px',
                            layout=self.tool_layout)

    def init_regress_ui(self):
        """线性拟合ui"""

        with self._init_tip_label_with_step_x(
                self._regress_analyse, u'线性拟合分析') as (widget_list, step_x):
            self.regress_step_x = step_x
            self.regress_mode_description = widgets.Textarea(
                value=u'1. 技术线最少拟合次数：\n'
                      u'检测至少多少次拟合曲线可以代表原始曲线y的走势，'
                      u'通过度量始y值和均线y_roll_mean的距离和原始y值和拟合回归的趋势曲线y_fit的距离的方法，默认使用metrics_rmse\n'
                      u'2. 技术线最优拟合次数：\n'
                      u'寻找多少次多项式拟合回归的趋势曲线可以完美的代表原始曲线y的走势\n'
                      u'3. 可视化技术线拟合曲线：\n'
                      u'通过步长参数在子金融序列中进行走势拟合，形成拟合曲线及上下拟合通道曲线，返回三条拟合曲线，组成拟合通道',
                disabled=False,
                layout=self.description_layout
            )
            widget_list.append(self.regress_mode_description)
            self.regress_mode = widgets.RadioButtons(
                options={u'技术线最少拟合次数': 0, u'技术线最优拟合次数': 1,
                         u'可视化技术线拟合曲线': 2},
                value=0,
                description=u'拟合模式:',
                disabled=False
            )
            widget_list.append(self.regress_mode)
        return widgets.VBox(widget_list,
                            # border='solid 1px',
                            layout=self.tool_layout)

    def init_golden_line_ui(self):
        """黄金分割ui"""
        with self._init_tip_label_with_step_x(
                self._golden_line_analyse, u'黄金分割分析', with_step_x=False) as (widget_list, _):
            self.golden_line_mode = widgets.RadioButtons(
                options={u'可视化黄金分隔带': 0, u'可视化黄金分隔带＋关键比例': 1,
                         u'可视化关键比例': 2},
                value=0,
                description=u'分隔模式:',
                disabled=False
            )
            widget_list.append(self.golden_line_mode)
            pt_tip_label = widgets.Label(u'比例设置仅对\'可视化关键比例\'生效', layout=self.label_layout)
            self.pt_range = widgets.FloatRangeSlider(
                value=[0.2, 0.8],
                min=0.1,
                max=0.9,
                step=0.1,
                description=u'比例设置:',
                disabled=False,
                continuous_update=False,
                orientation='horizontal',
                readout=True,
                readout_format='.1f',
            )
            pt_box = widgets.VBox([pt_tip_label, self.pt_range])
            widget_list.append(pt_box)

        return widgets.VBox(widget_list,
                            # border='solid 1px',
                            layout=self.tool_layout)

    def init_skeleton_ui(self):
        """价格骨架ui"""
        with self._init_tip_label_with_step_x(
                self._skeleton_analyse, u'价格骨架分析') as (widget_list, step_x):
            self.skeleton_step_x = step_x
            self.skeleton_mode = widgets.RadioButtons(
                options={u'骨架通道点位值': 200, u'三角骨架点位值': 100, u'最小值骨架点位值': 0,
                         u'最大值骨架点位值': 1, u'平均值骨架点位值': 2, u'中位数骨架点位值': 3,
                         u'最后元素骨架点位值': 4},
                value=200,
                description=u'骨架模式:',
                disabled=False
            )
            widget_list.append(self.skeleton_mode)
        return widgets.VBox(widget_list,
                            # border='solid 1px',
                            layout=self.tool_layout)

    def init_pair_speed_ui(self):
        """趋势敏感速度分析ui"""
        with self._init_tip_label_with_step_x(
                self._pair_speed_analyse, u'趋势敏感速度分析', with_step_x=False) as (widget_list, _):
            self.pair_speed_mode = widgets.RadioButtons(
                options={u'对比收盘敏感速度': 'close', u'对比涨跌敏感速度': 'p_change',
                         u'对比最高敏感速度': 'high', u'对比最低敏感速度': 'low'},
                value='close',
                description=u'对比模式:',
                disabled=False
            )
            widget_list.append(self.pair_speed_mode)
            resample_tip_label = widgets.Label(u'趋势敏感速度计算重采样周期', layout=self.label_layout)
            self.pair_resample = widgets.IntSlider(
                value=5,
                min=3,
                max=10,
                step=1,
                description=u'重采样',
                disabled=False,
                orientation='horizontal',
                readout=True,
                readout_format='d'
            )
            resample_box = widgets.VBox([resample_tip_label, self.pair_resample])
            widget_list.append(resample_box)
        return widgets.VBox(widget_list,
                            # border='solid 1px',
                            layout=self.tool_layout)

    def __init__(self, tool_set):
        """初始化技术分析界面"""
        super(WidgetTLTool, self).__init__(tool_set)

        rs_box = self.init_rs_ui()
        jump_box = self.init_jump_ui()
        pair_speed = self.init_pair_speed_ui()
        shift_distance = self.init_shift_distance_ui()
        regress = self.init_regress_ui()
        golden = self.init_golden_line_ui()
        skeleton = self.init_skeleton_ui()

        children = [rs_box, jump_box, pair_speed, shift_distance, regress, golden, skeleton]
        if self.scroll_factor_box:
            tl_box = widgets.Box(children,
                                 layout=self.scroll_widget_layout)
            # 需要再套一层VBox，不然外部的tab显示有问题
            self.widget = widgets.VBox([tl_box])
        else:
            # 一行显示两个，2个为一组，组装sub_children_group序列,
            sub_children_group = self._sub_children(children, len(children) / 2)
            sub_children_box = [widgets.HBox(sub_children) for sub_children in sub_children_group]
            self.widget = widgets.VBox(sub_children_box)

    def _pair_speed_analyse(self, bt):
        """趋势变化敏感速度分析action"""
        ABuProgress.clear_output()
        symbol1, symbol2 = self._choice_symbol_pair()
        start, end, n_folds = self._start_end_n_fold()
        speed_key = self.pair_speed_mode.value
        resample = self.pair_resample.value
        symbol1_speed, symbol2_speed, corr = calc_pair_speed(symbol1, symbol2, resample=resample, speed_key=speed_key,
                                                             start=start, end=end, n_folds=n_folds, show=True)
        if symbol1_speed is None:
            self.info_change_set_mode('{} and {}'.format(symbol1, symbol2))
        else:
            logging.info(u'{}趋势变化敏感速度{}'.format(symbol1, symbol1_speed))
            logging.info(u'{}趋势变化敏感速度{}'.format(symbol2, symbol2_speed))
            logging.info(u'{}与{}相关度{}'.format(symbol2, symbol2, corr))
            logging.info(u'{}与{}趋势相关敏感速度差{}'.format(symbol1, symbol2, (symbol1_speed - symbol2_speed) * corr))

    @single_fetch_symbol_analyse
    def _jump_line_analyse(self, kl, kl_tl, bt):
        """跳空缺口分析action"""
        # print('正在分析跳空缺口，请稍后...')
        if self.jump_mode.value == 0:
            jumps = calc_jump(kl, jump_diff_factor=self.jump_diff_factor.value)
        elif self.jump_mode.value == 1:
            jumps = calc_jump_line(kl, power_threshold=self.power_threshold.value,
                                   jump_diff_factor=self.jump_diff_factor.value)
        else:
            # 暂时固定加权比例为(0.5, 0.5)
            jumps = calc_jump_line_weight(kl, sw=(0.5, 0.5), power_threshold=self.power_threshold.value,
                                          jump_diff_factor=self.jump_diff_factor.value)
        display(jumps)

    @single_fetch_symbol_analyse
    def _rs_line_analyse(self, kl, kl_tl, bt):
        """支撑阻力线分析action"""
        if self.rs_mode.value == 0:
            # 只绘制支撑线
            kl_tl.show_support_trend(only_last=self.only_last.value, show=True, show_step=False)
        elif self.rs_mode.value == 1:
            # 只绘制阻力线
            kl_tl.show_resistance_trend(only_last=self.only_last.value, show=True, show_step=False)
        else:
            # 支撑线和阻力线都绘制
            kl_tl.show_support_resistance_trend(only_last=self.only_last.value, show=True, show_step=False)

    @single_fetch_symbol_analyse
    def _shift_distance_analyse(self, kl, kl_tl, bt):
        """位移路程比分析action"""
        kl_tl.show_shift_distance(how=EShiftDistanceHow(self.shift_distance_mode.value),
                                  step_x=self.shift_distance_step_x.value)

    @single_fetch_symbol_analyse
    def _regress_analyse(self, kl, kl_tl, bt):
        """走势线性回归分析action"""
        if self.regress_mode.value == 0:
            kl_tl.show_least_valid_poly()
        elif self.regress_mode.value == 1:
            kl_tl.show_best_poly()
        else:
            kl_tl.show_regress_trend_channel(step_x=self.regress_step_x.value)

    @single_fetch_symbol_analyse
    def _golden_line_analyse(self, kl, kl_tl, bt):
        """走势黄金分割分析action"""
        if self.golden_line_mode.value == 0:
            kl_tl.show_golden()
        elif self.golden_line_mode.value == 1:
            calc_golden(kl)
        else:
            kl_tl.show_percents(self.pt_range.value)

    @single_fetch_symbol_analyse
    def _skeleton_analyse(self, kl, kl_tl, bt):
        """走势骨架分析action"""
        step_x = self.skeleton_step_x.value
        skeleton_mode = self.skeleton_mode.value
        if skeleton_mode == 200:
            kl_tl.show_skeleton_channel(step_x=step_x)
        else:
            kl_tl.show_skeleton(how=ESkeletonHow(skeleton_mode), step_x=step_x)
